(* This file is distributed under the terms of the MIT License, also
   known as the X11 Licence.  A copy of this license is in the README
   file that accompanied the original distribution of this file.

   Based on code written by:
     Brian Aydemir
     Aaron Bohannon
     Arthur Charg\'eraud

   $Id: AssocList.v 2731 2009-02-24 17:29:24Z baydemir $ *)

(** A library for association lists, i.e., lists of pairs *)


Require Import Coq.Lists.List.
Require Import Coq.Logic.Decidable.
Require Import Coq.Classes.Equivalence.
Require Import Coq.Classes.EquivDec.

Require Import FSetsClass.
Require Import FSetsClassDecide.
Require Import FSetsClassFacts.


(* *********************************************************************** *)
(** * Implementation *)

(** We implement association lists as lists of key-value pairs.  We
    use [eq] as the equality on both keys and values.  This is evident
    primarily in the definition [binds].

    (Implementation note:  We can generalize the implementation to
    arbitrary equivalence relations.  For now, we have chosen not to,
    in order to keep the implementation simple.)

    Note that our library has an implicit convention for the "normal
    form" of an association list.  This normal form requires that a
    list be built only from [nil] (the empty list), [one] (the
    singleton list), and [app] (concatenation).  Additionally,
    concatenations should be associated to the right and the list
    should not contain any [nil] terms, unless it is the empty list
    itself.

    This allows association lists to be written in a slightly more
    uniform manner when compared to using both [cons] and [app] to
    build them.  The downsides are that Coq's [simpl] tactic will
    simplify instances of [one] down to [cons] and that there are
    instances in which one needs to write association lists that are
    not in normal form (e.g., some concatenation will need to be
    associated to the left).  The [simpl_alist] and [rewrite_alist]
    tactics below are designed to minimize the impact of these
    downsides.  We use a [Strategy] declaration at the end of this
    file to prevent Coq's tactics from simplifying [one].

    We have considered other alternative implementations, but the scheme
    described above seems to work the best in practice. *)

(*  >> Purposefully not including this in generated HTML documentation. <<

    Implementation notes, from when Coq did not support [Strategy]
    declarations:

    There are various mechanisms for enforcing that [one] is never
    simplified down to [cons], e.g., by making its definition
    abstract.  However, this is undesirable because it interacts
    poorly with Coq's [inversion] tactic.  For example, in order to
    sensibly invert the conclusion of [uniq_push], Coq needs to reduce
    [((x ~ a) ++ E)] down to a term with a constructor at its head.
    This is impossible if Coq cannot unfold the definition of [one].

    Alternatively, we could could define a completely new inductive
    datatype for association lists that had three constructors---one
    each corresponding to [nil], [one], and [app].  Associativity of
    [app] then becomes a problem because we need to define a suitable
    equality relation on association lists, instead of simply being
    able to use [eq]. *)


(* ********************************************************************* *)
(** * Basic definitions *)

(** Implicit arguments are enabled for the following definitions. *)

Set Implicit Arguments.

(** [one] constructs an association list consisting of exactly one
    binding.  We define an infix notation for it and ensure that the
    arguments to [app] are interpreted in the right scope, i.e.,
    [list_scope].

    Implementation note: The levels associated with the notation are
    determined by an unrelated (but similar) notation in Coq's BinPos
    library and the notation for [app] in Coq's List library. *)

Definition one (C : Type) (item : C) : list C := cons item nil.

Notation "x ~ a" :=
  (one (x, a))
  (at level 7, a at level 59, left associativity)
  : list_scope.

Arguments Scope app [ type_scope list_scope list_scope ].

(** [dom] computes the domain of an association list, i.e., the
    set consisting of its keys. *)

Fixpoint dom (key C : Type) `{FSet key} (E : list (key*C)) {struct E}
  : fset key :=
  match E with
    | nil => empty
    | (x, _) :: E' => add x (dom E')
  end.

(** [binds] is a ternary predicate that holds when a key-value pair
    appears somewhere in the given association list. *)

Definition binds
  (key A : Type) (x : key) (a : A) (E : list (key*A))
  : Prop :=
  List.In (x, a) E.

(** [disjoint] is a binary predicate that holds when the domains of
    two association lists are disjoint. *)

Definition disjoint
  (key A B : Type) `{FSet key} (E : list (key*A)) (F : list (key*B))
  : Prop :=
  dom E `inter` dom F [<=] empty.

(** [map] applies a function to each of the values in an association
    list. *)

Definition map
  (key A B : Type) (f : A -> B) (E : list (key*A))
  : list (key*B) :=
  List.map (fun b => match b with (x, a) => (x, f a) end) E.

(** [uniq] is unary predicate that holds if and only if each key is
    bound at most once in the given association list.  Note that
    [uniq] is defined in terms of [one], not [cons]. *)

Inductive uniq (key A : Type) `{Fset : FSet key} : list (key*A) -> Prop :=
  | uniq_nil :
      uniq nil
  | uniq_push : forall x a E,
      uniq E ->
      x `notin` dom E ->
      uniq (x ~ a ++ E).

(** Unless stated otherwise, in the remainder of this file, implicit
    arguments will not be declared by default. *)

Unset Implicit Arguments.


(* ********************************************************************* *)
(** * List properties *)

(** The following properties are used mainly for rewriting association
    lists into the normal form described above.  See the [simpl_alist]
    and [rewrite_alist] tactics below. *)

Section ListProperties.
  Variable  X : Type.
  Variables x y : X.
  Variables l l1 l2 l3 : list X.

  Lemma cons_app_one :
    cons x l = one x ++ l.
  Proof. clear. reflexivity. Qed.

  Lemma cons_app_assoc :
    (cons x l1) ++ l2 = cons x (l1 ++ l2).
  Proof. clear. reflexivity. Qed.

  Lemma app_assoc :
    (l1 ++ l2) ++ l3 = l1 ++ (l2 ++ l3).
  Proof. clear. auto with datatypes. Qed.

  Lemma app_nil_1 :
    nil ++ l = l.
  Proof. clear. reflexivity. Qed.

  Lemma app_nil_2 :
    l ++ nil = l.
  Proof. clear. auto with datatypes. Qed.

  Lemma in_nil :
    List.In x nil <-> False.
  Proof. clear. split; inversion 1. Qed.

  Lemma in_one :
    List.In x (one y) <-> x = y.
  Proof. clear. split. inversion 1; intuition. constructor; intuition. Qed.

  Lemma in_app :
    List.In x (l1 ++ l2) <-> List.In x l1 \/ List.In x l2.
  Proof. clear. split; auto using List.in_or_app, List.in_app_or. Qed.

End ListProperties.


(* ********************************************************************* *)
(** * Properties of [map] and [dom] *)

(** The following lemmas are used mainly to simplify applications of
    [map] and [dom] to association lists.  See also the [simpl_alist]
    and [rewrite_alist] tactics below. *)

Section Properties.
  Variables A B key : Type.
  Variable  f       : A -> B.
  Variable  x       : key.
  Variable  b       : A.
  Variables E F G   : list (key*A).
  Variable  R       : relation key.
  Variable  Equiv   : @Equivalence key R.
  Variable  Eqdec   : @EqDec key R Equiv.
  Variable  Fset    : @FSet key R Equiv Eqdec.

  Lemma map_nil :
    map f (@nil (key*A)) = nil.
  Proof. clear. reflexivity. Qed.

  Lemma map_cons :
    map f ((x, b) :: E) = (x, f b) :: map f E.
  Proof. clear. reflexivity. Qed.

  Lemma map_one :
    map f (x ~ b) = (x ~ f b).
  Proof. clear. reflexivity. Qed.

  Lemma map_app :
    map f (E ++ F) = map f E ++ map f F.
  Proof. clear. unfold map. rewrite List.map_app. reflexivity. Qed.

  Lemma dom_nil :
    dom (@nil (key*A)) = empty.
  Proof. clear. reflexivity. Qed.

  Lemma dom_cons :
    dom ((x, b) :: E) [=] singleton x `union` dom E.
  Proof. clear. simpl. fsetdec key. Qed.

  Lemma dom_one :
    dom (x ~ b) [=] singleton x.
  Proof. clear. simpl. fsetdec key. Qed.

  Lemma dom_app :
    dom (E ++ F) [=] dom E `union` dom F.
  Proof. clear. induction E as [ | [x1 a1] ]; simpl; fsetdec key. Qed.

  Lemma dom_map :
    dom (map f E) [=] dom E.
  Proof. clear. induction E as [ | [x1 a1] ]; simpl; fsetdec key. Qed.

End Properties.


(* ********************************************************************* *)
(** * The [simpl_alist] tactic *)

(** Rewriting hints. *)

Hint Rewrite cons_app_one cons_app_assoc   : rewr_list.
Hint Rewrite app_assoc app_nil_1 app_nil_2 : rewr_list.
Hint Rewrite in_nil in_one in_app          : rewr_list_in.

Hint Rewrite map_nil map_cons map_one map_app         : rewr_map.
Hint Rewrite dom_nil dom_cons dom_one dom_app dom_map : rewr_dom.

(** The [simpl_alist] tactic rewrites association lists so that they
    are in the normal form described above.  Similar to the [simpl]
    tactic, we define "[in *]" and "[in H]" variants of the tactic. *)

Ltac simpl_alist :=
  autorewrite with rewr_list rewr_list_in rewr_map rewr_dom.
Tactic Notation "simpl_alist" "in" hyp(H) :=
  autorewrite with rewr_list rewr_list_in rewr_map rewr_dom in H.
Tactic Notation "simpl_alist" "in" "*" :=
  autorewrite with rewr_list rewr_list_in rewr_map rewr_dom in *.


(* ********************************************************************* *)
(** * The [rewrite_alist] tactic *)

(** The tactic [(rewrite_alist E)] replaces an association list in
    the conclusion of the goal with [E].  Suitability for
    replacement is determined by whether [simpl_alist] can put [E]
    and the chosen environment in the same normal form, up to
    convertibility's in Coq.  We also define an "[in H]" variant
    that performs the replacement in a hypothesis [H].  *)

Tactic Notation "rewrite_alist" constr(E) :=
  match goal with
    | |- context[?x] =>
      change x with E
    | |- context[?x] =>
      replace x with E;
        [ | try reflexivity; simpl_alist; reflexivity ]
  end.

Tactic Notation "rewrite_alist" constr(E) "in" hyp(H) :=
  match type of H with
    | context[?x] =>
      change x with E in H
    | context[?x] =>
      replace x with E in H;
        [ | try reflexivity; simpl_alist; reflexivity ]
  end.


(* ********************************************************************* *)
(** * Basic facts about [disjoint] *)

Section Disjoint.
  Implicit Types A B C : Type.

  Lemma disjoint_sym_1 :
    forall A B `{FSet key} (E : list (key*A)) (F : list (key*B)),
    disjoint E F ->
    disjoint F E.
  Proof. unfold disjoint. fsetdec key. Qed.

  Lemma disjoint_sym :
    forall A B `{FSet key} (E : list (key*A)) (F : list (key*B)),
    disjoint E F <-> disjoint F E.
  Proof. intuition auto using disjoint_sym_1. Qed.

  Lemma disjoint_one_1 :
    forall A B `{FSet key} (x : key) (a : A) (F : list (key*B)),
    disjoint (x ~ a) F ->
    x `notin` dom F.
  Proof. unfold disjoint. simpl. intros. fsetdec key. Qed.

  Lemma disjoint_one_2 :
    forall A B `{FSet key} (x : key) (a : A) (F : list (key*B)),
    x `notin` dom F ->
    disjoint (x ~ a) F.
  Proof. unfold disjoint. simpl. intros. fsetdec key. Qed.

  Lemma disjoint_one_l :
    forall A B `{FSet key} (x : key) (a : A) (E : list (key*B)),
    disjoint (x ~ a) E <-> x `notin` dom E.
  Proof. unfold disjoint. intros. simpl. split; fsetdec key. Qed.

  Lemma disjoint_one_r :
    forall A B `{FSet key} (x : key) (a : A) (E : list (key*B)),
    disjoint E (x ~ a) <-> x `notin` dom E.
  Proof. intros. rewrite disjoint_sym. apply disjoint_one_l. Qed.

  Lemma disjoint_app_1 :
    forall A B `{FSet key} (E F : list (key*A)) (G : list (key*B)),
    disjoint (E ++ F) G ->
    disjoint E G.
  Proof. unfold disjoint. intros. simpl_alist in *. fsetdec key. Qed.

  Lemma disjoint_app_2 :
    forall A B `{FSet key} (E F : list (key*A)) (G : list (key*B)),
    disjoint (E ++ F) G ->
    disjoint F G.
  Proof. unfold disjoint. intros. simpl_alist in *. fsetdec key. Qed.

  Lemma disjoint_app_3 :
    forall A B `{FSet key} (E F : list (key*A)) (G : list (key*B)),
    disjoint E G ->
    disjoint F G ->
    disjoint (E ++ F) G.
  Proof. unfold disjoint. intros. simpl_alist in *. fsetdec key. Qed.

  Lemma disjoint_app_l :
    forall A B `{FSet key} (E F : list (key*A)) (G : list (key*B)),
    disjoint (E ++ F) G <-> disjoint E G /\ disjoint F G.
  Proof.
    intros; intuition eauto 2 using
      disjoint_app_1, disjoint_app_2, disjoint_app_3.
  Qed.

  Lemma disjoint_app_r :
    forall A B `{FSet key} (E F : list (key*A)) (G : list (key*B)),
    disjoint G (E ++ F) <-> disjoint E G /\ disjoint F G.
  Proof. intros. rewrite disjoint_sym. apply disjoint_app_l. Qed.

  Lemma disjoint_map_1 :
    forall A B C `{FSet key} (E : list (key*A)) (F : list (key*B)) (f:A->C),
    disjoint (map f E) F ->
    disjoint E F.
  Proof. unfold disjoint. intros. simpl_alist in *. fsetdec key. Qed.

  Lemma disjoint_map_2 :
    forall A B C `{FSet key} (E : list (key*A)) (F : list (key*B)) (f:A->C),
    disjoint E F ->
    disjoint (map f E) F.
  Proof. unfold disjoint. intros. simpl_alist in *. fsetdec key. Qed.

  Lemma disjoint_map_l :
    forall A B C `{FSet key} (E : list (key*A)) (F : list (key*B)) (f:A->C),
    disjoint (map f E) F <-> disjoint E F.
  Proof. intros; intuition eauto using disjoint_map_1, disjoint_map_2. Qed.

  Lemma disjoint_map_r :
    forall A B C `{FSet key} (E : list (key*A)) (F : list (key*B)) (f:A->C),
    disjoint F (map f E) <-> disjoint E F.
  Proof. intros. rewrite disjoint_sym. apply disjoint_map_l. Qed.

End Disjoint.


(* ********************************************************************* *)
(** * Basic facts about [uniq] *)

(** The following lemmas are facts about [uniq] with respect to the
    basic functions ([one], [app], and [map]) that can be used to
    build association lists. *)

Section UniqProperties.
  Variables A B key : Type.
  Variables f       : A -> B.
  Variables x       : key.
  Variables a b     : A.
  Variables E F G   : list (key*A).
  Variable  R       : relation key.
  Variable  Equiv   : @Equivalence key R.
  Variable  Eqdec   : @EqDec key R Equiv.
  Variable  Fset    : @FSet key R Equiv Eqdec.

  Lemma uniq_one :
    uniq (x ~ b).
  Proof.
    clear. rewrite_alist ((x ~ b) ++ nil).
    apply uniq_push. apply uniq_nil. apply empty_intro.
  Qed.

  Lemma uniq_app_1 :
    uniq (E ++ F) -> uniq E.
  Proof.
    clear. intros H; induction E as [ | [x1 a1] E' ].
      apply uniq_nil.
      inversion H; subst. simpl_alist in *. apply uniq_push.
        auto.
        simpl_alist in *. fsetdec key.
  Qed.

  Lemma uniq_app_2 :
    uniq (E ++ F) -> uniq F.
  Proof.
    clear. intros H; induction E as [ | [x1 a1] E'].
      apply H.
      inversion H; subst. auto.
  Qed.

  Lemma uniq_app_3 :
    uniq (E ++ F) -> disjoint E F.
  Proof.
    clear. intros H. red. induction E as [ | [x1 a1] E' IHE' ].
      fsetdec key.
      inversion H; subst. simpl_alist in *. lapply IHE'.
        fsetdec key.
        assumption.
  Qed.

  Lemma uniq_app_4 :
    uniq E ->
    uniq F ->
    disjoint E F ->
    uniq (E ++ F).
  Proof.
    clear. intros HE HF Hd. induction E as [ | [x1 a1] E']; simpl_alist in *.
      assumption.
      rewrite disjoint_app_l in Hd. inversion HE; subst. apply uniq_push.
          intuition.
          rewrite disjoint_one_l in Hd. simpl_alist. fsetdec key.
  Qed.

  Lemma uniq_app :
    uniq (E ++ F) <-> uniq E /\ uniq F /\ disjoint E F.
  Proof.
    clear; intuition eauto using
      uniq_app_1, uniq_app_2, uniq_app_3, uniq_app_4.
  Qed.

  Lemma uniq_map_1 :
    uniq (map f E) ->
    uniq E.
  Proof.
    clear. intros H. induction E as [ | [x1 a1] E']; simpl_alist in *.
      apply uniq_nil.
      inversion H; subst. apply uniq_push; simpl_alist in *; auto.
  Qed.

  Lemma uniq_map_2 :
    uniq E ->
    uniq (map f E).
  Proof.
    clear. intros H. induction E as [ | [x1 a1] E']; simpl_alist in *.
      apply uniq_nil.
      inversion H; subst. apply uniq_push; simpl_alist in *; auto.
  Qed.

  Lemma uniq_map :
    uniq (map f E) <-> uniq E.
  Proof. clear. intuition eauto using uniq_map_1, uniq_map_2. Qed.

End UniqProperties.


(* ********************************************************************* *)
(** * Basic facts about [binds] *)

(** The following lemmas are facts about [binds] with respect to the
    basic functions ([one] and [app]) that can be used to build
    association lists. *)

Section BindsProperties.
  Variable  A B key : Type.
  Variables f       : A -> B.
  Variables x y     : key.
  Variables a b     : A.
  Variables E F G   : list (key*A).
  Variable  R       : relation key.
  Variable  Equiv   : @Equivalence key R.
  Variable  Eqdec   : @EqDec key R Equiv.
  Variable  Fset    : @FSet key R Equiv Eqdec.

  Lemma binds_nil :
    binds x a nil <-> False.
  Proof. clear. split. inversion 1. intuition. Qed.

  Lemma binds_one_1 :
    x = y ->
    a = b ->
    binds x a (y ~ b).
  Proof. clear. intros. red. simpl. left. congruence. Qed.

  Lemma binds_one_2 :
    binds x a (y ~ b) ->
    x = y.
  Proof. clear. intros H1. inversion H1; intuition congruence. Qed.

  Lemma binds_one_3 :
    binds x a (y ~ b) ->
    a = b.
  Proof. clear. intros H1. inversion H1; intuition congruence. Qed.

  Lemma binds_one :
    binds x a (y ~ b) <-> x = y /\ a = b.
  Proof.
    clear. intuition eauto using binds_one_1, binds_one_2, binds_one_3.
  Qed.

  Lemma binds_app_1 :
    binds x a E ->
    binds x a (E ++ F).
  Proof. clear. intros H. red in H |- *. rewrite in_app. auto. Qed.

  Lemma binds_app_2 :
    binds x a F ->
    binds x a (E ++ F).
  Proof. clear. intros H. red in H |- *. rewrite in_app. auto. Qed.

  Lemma binds_app_3 :
    binds x a (E ++ F) ->
    binds x a E \/ binds x a F.
  Proof.
    clear. induction E as [ | [x1 a1] E' IHE' ].
      auto.
      unfold binds in *. simpl. intros [H | H].
        injection H; intros; subst. auto.
        apply IHE' in H. intuition.
  Qed.

  Lemma binds_app :
    binds x a (E ++ F) <-> binds x a E \/ binds x a F.
  Proof.
    clear. intuition auto using binds_app_1, binds_app_2, binds_app_3.
  Qed.

  Lemma binds_dom_contradiction : forall (E : list (key*A)),
    binds x a E ->
    x `notin` dom E ->
    False.
  Proof.
    clear. induction E as [ | [x1 a1] E' IHE' ].
      inversion 1.
      unfold binds in *. simpl. intros [H | H] J.
        injection H. intros. fsetdec key.
        apply IHE' in H. trivial. fsetdec key.
  Qed.

  Lemma binds_app_uniq :
    uniq (E ++ F) ->
    (binds x a (E ++ F) <->
      (binds x a E /\ x `notin` dom F) \/
      (binds x a F /\ x `notin` dom E)).
  Proof.
    clear. intros H1.
    rewrite uniq_app in H1. unfold disjoint in H1.
    assert (H : x `notin` dom F \/ x `notin` dom E) by fsetdec key.
    rewrite binds_app.
    intuition eauto using binds_dom_contradiction.
  Qed.

  Lemma binds_map :
    binds x a E ->
    binds x (f a) (map f E).
  Proof.
    clear. induction E as [ | [x1 a1] E' ]; simpl; intros H.
      inversion H.
      unfold binds in *. simpl in *. inversion H.
        left. congruence.
        right. auto.
  Qed.

  Lemma binds_map_inv :
    (forall a b, f a = f b -> a = b) ->
    binds x (f a) (map f E) ->
    binds x a E.
  Proof.
    clear. induction E as [ | [x1 a1] E' ]; simpl; intros J H.
      inversion H.
      unfold binds in *. simpl in *. inversion H as [K | K].
        left. injection K. intros. f_equal; auto.
        right. auto.
  Qed.

End BindsProperties.


(* ********************************************************************* *)
(** * Hints *)

Hint Resolve app_assoc app_nil_1 app_nil_2   : alist.
Hint Resolve map_nil map_one map_app         : alist.
Hint Resolve dom_nil dom_one dom_app dom_map : alist.

Hint Resolve uniq_nil uniq_one uniq_push : alist.
Hint Resolve uniq_app_4 uniq_map_2       : alist.

Hint Resolve binds_one_1               : alist.
Hint Resolve binds_app_1 binds_app_2   : alist.
Hint Resolve binds_map                 : alist.
Hint Immediate binds_dom_contradiction : alist.


(* ********************************************************************* *)
(** * Preventing reduction *)

Strategy opaque [ one ].
