(* This file is distributed under the terms of the MIT License, also
   known as the X11 Licence.  A copy of this license is in the README
   file that accompanied the original distribution of this file.

   Based on code written by:
     Brian Aydemir
     Aaron Bohannon

   $Id: SwapClass.v 2731 2009-02-24 17:29:24Z baydemir $ *)

Require Import Coq.Classes.Equivalence.
Require Import Coq.Classes.EquivDec.
Require Import Coq.Lists.List.

Require Import Atom.
Require Import DefaultSolver.


(* *********************************************************************** *)
(** * Class definition *)

Class Swap (A T : Type) `(Atom A) := {
  swap         : (A * A) -> T -> T;
  swap_same    : forall a x, swap (a, a) x = x;
  swap_invol   : forall ab x, swap ab (swap ab x) = x;
  swap_distrib : forall ab c d x,
    swap ab (swap (c, d) x) =
    swap (swap_atom ab c, swap_atom ab d) (swap ab x)
}.


(* *********************************************************************** *)
(** * Tactic support *)

(** The following hints are especially useful for defining instances
    of [Swap]. *)

Hint Resolve @swap_same @swap_invol : swap_distrib swap.
Hint Resolve @swap_distrib          : swap_distrib.

(** Turn applications of [f] into applications of [swap]. *)

Ltac swapify f :=
  repeat (match goal with
            | H : context [f ?ab ?x] |- _ =>
              progress change (f ab x) with (swap ab x) in H
            | |- context [f ?ab ?x] =>
              progress change (f ab x) with (swap ab x)
          end).

(** Turn applications of [swap] into applications of [swap_atom]. *)

Ltac swapify_atom :=
  repeat (match goal with
            | H : context [swap ?ab ?x] |- _ =>
              progress change (swap ab x) with (swap_atom ab x) in H
            | |- context [swap ?ab ?x] =>
              progress change (swap ab x) with (swap_atom ab x)
          end).


(* *********************************************************************** *)
(** * Class instances *)

(** Swapping on atoms. *)

Instance Swap_Atom `{Atom A} : Swap A A _ := {
  swap := swap_atom
}.
Proof.
  apply swap_atom_same.
  apply swap_atom_invol.
  apply swap_atom_distrib.
Defined.

(** Swapping on pairs. *)

Definition swap_pair
  `{X : Atom A, ! Swap A T1 X, ! Swap A T2 X} (ab : A * A) (xy : T1 * T2)
  : T1 * T2 :=
  let (x, y) := xy in (swap ab x, swap ab y).

Instance Swap_pair
  `{X : Atom A, ! Swap A T1 X, ! Swap A T2 X}
  : Swap A (T1 * T2) _ := {
  swap := swap_pair
}.
Proof.
  induction x; default_simp; auto with swap swap_distrib.
  induction x; default_simp; auto with swap swap_distrib.
  induction x; default_simp; auto with swap swap_distrib.
Defined.

(** Swapping on lists. *)

Fixpoint swap_list `{Swap A T} (ab : A * A) (xs : list T) : list T :=
  match xs with
    | nil => nil
    | y :: ys => (swap ab y) :: (swap_list ab ys)
  end.

Instance Swap_list `{Swap A T} : Swap A (list T) _ := {
  swap := swap_list
}.
Proof.
  induction x; default_simp; auto with swap swap_distrib.
  induction x; default_simp; auto with swap swap_distrib.
  induction x; default_simp; auto with swap swap_distrib.
Defined.


(* *********************************************************************** *)
(** * Properties *)

Lemma swap_injective : forall `{Swap A T} (ab : A * A) (x y : T),
  swap ab x = swap ab y ->
  x = y.
Proof.
  intros A T ? ? ? ? ab x y H.
  rewrite <- (swap_invol ab x).
  rewrite <- (swap_invol ab y).
  f_equal; trivial.
Qed.
