{- | This module defines the functions necessary to parse an Ott input
   file and check that it is well formed.

   Implementation notes (BEA):

   1. The comments below assume some level of familiarity with Parsec.

   2. Parsers below are implemented assuming that they don't need to
      consume any initial whitespace.  This means that they all need
      to be implemented as lexeme parsers, ones that consume trailing
      whitespace.  A few parsers are not lexeme parsers; these are
      noted below.
-}

module Parser ( parseOttFile ) where

import Data.Maybe ( catMaybes )
import Text.ParserCombinators.Parsec
import Text.ParserCombinators.Parsec.Language ( emptyDef )
import qualified Text.ParserCombinators.Parsec.Token as P

import AST
import MyLibrary ( getResult, manyTill1, sepStrings )


{- ----------------------------------------------------------------------- -}
{- * Basic, language specific parsers -}

{- | A parser \"module\" used mainly for handling whitespace,
   comments, and reserved words in Ott input files. -}

ottTokenParser :: P.TokenParser st
ottTokenParser = P.makeTokenParser (emptyDef { P.commentLine = "%" })

{- | A parser for whitespace, including comments. -}

whiteSpace :: CharParser st ()
whiteSpace = P.whiteSpace ottTokenParser

{- | Constructs a lexeme parser from its argument. -}

lexeme :: CharParser st a -> CharParser st a
lexeme = P.lexeme ottTokenParser

{- | A lexeme parser for identifiers. -}

identifier :: CharParser st String
identifier = P.identifier ottTokenParser

{- | A lexeme parser that parses a natural number, i.e., a non-empty
   sequence of digits.  It does not consume any input on failure. -}

natural :: CharParser st String
natural = lexeme (try (many1 digit) <?> "natural nunmber")

{- | A lexeme parser for reserved words. -}

reserved :: String -> CharParser st ()
reserved = P.reserved ottTokenParser

{- | A lexeme parser that is otherwise similar to the 'string' parser.
   It does not consume any input on failure. -}

text :: String -> CharParser st String
text str = lexeme $ try $ string str


{- ----------------------------------------------------------------------- -}
{- * Parsing names -}

{- | A variable (e.g., metavariable, nonterminal, etc.) consists of a
   root and suffix followed by at least one whitespace character. -}

var :: Parser UnknownSym
var = lexeme $ do { pos <- getPosition
                  ; r   <- root
                  ; s   <- suffix
                  ; space
                  ; return $ UnknownSym pos r s
                  }

{- | The root of a variable consists of a non-empty sequence of
   letters and underscores.  This is NOT a lexeme parser. -}

root :: Parser Root
root = many1 ((letter <|> char '_') <?> "") <?> "name"

{- | A suffix of a variable is a (possibly empty) sequence of digits
   and primes.  This is NOT a lexeme parser. -}

suffix :: Parser Suffix
suffix = many (digit <|> char '\'')

{- | 'name' is a lexeme parser version of 'root'. -}

name :: Parser Root
name = lexeme root


{- ----------------------------------------------------------------------- -}
{- * Ignoring homomorphisms -}

{- | Consumes a homomorphism from the input stream. -}

ignoreHom :: Parser ()
ignoreHom =
    do { text "{{" <?> "homomorphism \"{{ ... }}\""
       ; anyChar `manyTill` text "}}"
       ; return ()
       }


{- ----------------------------------------------------------------------- -}
{- * Parsing metavariable declarations -}

{- | A metavariable declaration consists of a non-empty list of
   metavariable roots and a possibly empty list of homomorphisms -}

metavarDecl :: Parser MetavarDecl
metavarDecl =
    do { pos <- getPosition
       ; reserved "metavar" <?> "metavariable declaration \"metavar ...\""
       ; names <- name `sepBy1` (text "," <?> "\",\" and another name")
       ; text "::="
       ; homs <- many metavarHom
       ; return (MetavarDecl pos names (catMaybes homs ++ defaults))
       }
    where
      defaults = [ CoqMvImplHom "string" ]

      metavarHom =
          do { text "{{" <?> "homomorphism \"{{ ... }}\""
             ; do { reserved "coq"
                  ; strs <- many1 (identifier <|> natural)
                  ; text "}}"
                  ; return $ Just $ CoqMvImplHom (sepStrings " " strs)
                  }
               <|>
               do { reserved "phantom"
                  ; text "}}"
                  ; return $ Just $ MvPhantom
                  }
               <|>
               do { anyChar `manyTill` text "}}"
                  ; return Nothing
                  }
             }


{- ----------------------------------------------------------------------- -}
{- * Parsing definitions of nonterminals -}

{- | A rule defines a nonterminal.  The definition begins with a list
   of roots, followed by a "name" and homomorphisms.  The remainder of
   the rule is a sequence of productions. -}

rule :: Parser PreRule
rule =
    do { pos <- getPosition
       ; es <- name `sepBy1` (text "," <?> "\",\" and another name")
       ; text "::"
       ; n <- ruleName
       ; text "::="
       ; many ignoreHom
       ; prods <- many production
       ; return $ Rule pos es n prods
       }
    where
      ruleName =
          lexeme (try (many1 (letter <|> char '_' <?> "")) <?> "name")
          <|>
          -- BEA: Seems like unnecessary flexbility in the input language.
          lexeme (do { char '\''
                     ; (letter <|> char '_') `manyTill` char '\''
                     })

{- | Each production in a rule starts with a \"|\" followed by a
   non-empty sequence of elements.  This is followed by flags, a name
   for this production, binding specifications, and homomorphisms. -}

production :: Parser PreProduction
production =
    do { pos <- getPosition
       ; text "|"
       ; es <- element `manyTill1` text "::"
       ; flag <- many productionFlag
       ; text "::"
       ; constr <- identifier
       ; bs <- many bindingSpec
       ; many ignoreHom
       ; return $ Production pos es flag constr bs
       }
    where
      productionFlag =
          do { reserved "I"; return $ IFlag }
          <|>
          do { reserved "M"; return $ MFlag }
          <|>
          do { reserved "S"; return $ SFlag }

{- | An element is a terminal, nonterminal, or metavariable.

   Implementation note (BEA): The difference between these three is
   sorted out later since we don't have enough information at this
   stage to make this distinction.  We are careful to identify
   potential nonterminals and metavariables as such. -}

element :: Parser UnknownElement
element =
    try (do { v <- var
            ; return $ Variable v
            })
    <|>
    try (do { s <- noneOf " \v\f\t\r\n" `manyTill1` whiteSpace
            ; return $ Unknown s
            })

{- | A binding specification helps define the binding structure of a
   given production.  Similar to 'element', we sort out references to
   variables later. -}

bindingSpec :: Parser PreBindingSpec
bindingSpec =
    do { pos <- getPosition
       ; text "(+"
       ; b <- do { reserved "bind"
                 ; v1 <- var
                 ; reserved "in"
                 ; v2 <- var
                 ; return $ BindDecl pos v1 v2
                 }
       ; text "+)"
       ; return b
       }


{- ----------------------------------------------------------------------- -}
{- * Parsing an entire file -}

{- | A parser for an entire Ott file. -}

ottFileParser :: Parser PreAST
ottFileParser = do { whiteSpace
                   ; mvds <- many metavarDecl
                   ; reserved "grammar"
                   ; rs <- many rule
                   ; return $ PreAST mvds rs
                   }

{- | Parses an entire Ott file given its name. -}

parseOttFile :: String -> IO PreAST
parseOttFile filename =
    do { answer <- parseFromFile ottFileParser filename
       ; return (getResult answer)
       }
