Add LoadPath "metatheory-8.2".
Require Export Coq.Program.Equality.
Require Export Metatheory.
Require Export LNgenFacts.
Require Import Wf_nat.

(** NOTE: Auxiliary theorems are hidden in generated documentation.
    In general, there is a [_rec] version of every lemma involving
    [open] and [close]. *)

(** For generated HTML documentation. *)

(** remove printing <= *)

(** For proving the uniqueness of [lc] proofs. *)

Axiom axiom_fun_ext_eq :
  forall (A : Type) (B : A -> Type) (f g : forall x : A, B x),
    (forall x, f x = g x) ->
    f = g.


(* *********************************************************************** *)
(** * Metavariable declarations *)

(** The rest of this file assumes that [EqDec] and [Atom] instances
    are already declared for the implementation types below. *)

Notation another_atom := (string).

Notation atom := (string).


(* *********************************************************************** *)
(** * Nonterminal declarations *)

Inductive exp : Set :=
  | var : atom -> exp
  | var_ind : nat -> exp
  | abs : exp -> exp
  | app : exp -> exp -> exp
  | xvar : another_atom -> exp
  | xvar_ind : nat -> exp
  | xabs : exp -> exp.

Scheme exp_ind' := Induction for exp Sort Prop.

Definition exp_mutind :=
  fun H H1 H2 H3 H4 H5 H6 H7 =>
  exp_ind' H H1 H2 H3 H4 H5 H6 H7.

Scheme exp_rec' := Induction for exp Sort Set.

Definition exp_mutrec :=
  fun H H1 H2 H3 H4 H5 H6 H7 =>
  exp_rec' H H1 H2 H3 H4 H5 H6 H7.


(* *********************************************************************** *)
(** * Close *)

Fixpoint close_eX_rec (n : nat) (X : another_atom) (e : exp) {struct e} : exp :=
  match e with
    | var x => var x
    | var_ind n1 => var_ind n1
    | abs e1 => abs (close_eX_rec n X e1)
    | app e1 e2 => app (close_eX_rec n X e1) (close_eX_rec n X e2)
    | xvar X1 => if (X == X1) then (xvar_ind n) else (xvar X1)
    | xvar_ind n1 => if (lt_ge_dec n1 n) then (xvar_ind n1) else (xvar_ind (S n1))
    | xabs e1 => xabs (close_eX_rec (S n) X e1)
  end.

Fixpoint close_ex_rec (n : nat) (x : atom) (e : exp) {struct e} : exp :=
  match e with
    | var x1 => if (x == x1) then (var_ind n) else (var x1)
    | var_ind n1 => if (lt_ge_dec n1 n) then (var_ind n1) else (var_ind (S n1))
    | abs e1 => abs (close_ex_rec (S n) x e1)
    | app e1 e2 => app (close_ex_rec n x e1) (close_ex_rec n x e2)
    | xvar X => xvar X
    | xvar_ind n1 => xvar_ind n1
    | xabs e1 => xabs (close_ex_rec n x e1)
  end.

Definition close_eX X e := close_eX_rec 0 X e.

Definition close_ex x e := close_ex_rec 0 x e.


(* *********************************************************************** *)
(** * Free variables *)

Fixpoint fv_eX (e : exp) {struct e} : fset another_atom :=
  match e with
    | var x => empty
    | var_ind n => empty
    | abs e1 => (fv_eX e1)
    | app e1 e2 => (fv_eX e1) `union` (fv_eX e2)
    | xvar X => (singleton X)
    | xvar_ind n => empty
    | xabs e1 => (fv_eX e1)
  end.

Fixpoint fv_ex (e : exp) {struct e} : fset atom :=
  match e with
    | var x => (singleton x)
    | var_ind n => empty
    | abs e1 => (fv_ex e1)
    | app e1 e2 => (fv_ex e1) `union` (fv_ex e2)
    | xvar X => empty
    | xvar_ind n => empty
    | xabs e1 => (fv_ex e1)
  end.


(* *********************************************************************** *)
(** * Open *)

(** The definitions below are only correct under the following
    condition: when substituting for index [n], the [degree] of
    the term being substituted in should also be [n]. *)

Fixpoint open_eX_rec (n : nat) (e : exp) (e1 : exp) {struct e1} : exp :=
  match e1 with
    | var x => var x
    | var_ind n1 => var_ind n1
    | abs e2 => abs (open_eX_rec n e e2)
    | app e2 e3 => app (open_eX_rec n e e2) (open_eX_rec n e e3)
    | xvar X => xvar X
    | xvar_ind n1 =>
      match lt_eq_lt_dec n1 n with
        | inleft (left _)  => xvar_ind n1
        | inleft (right _) => e
        | inright _        => xvar_ind (n1 - 1)
      end
    | xabs e2 => xabs (open_eX_rec (S n) e e2)
  end.

Fixpoint open_ex_rec (n : nat) (e : exp) (e1 : exp) {struct e1} : exp :=
  match e1 with
    | var x => var x
    | var_ind n1 =>
      match lt_eq_lt_dec n1 n with
        | inleft (left _)  => var_ind n1
        | inleft (right _) => e
        | inright _        => var_ind (n1 - 1)
      end
    | abs e2 => abs (open_ex_rec (S n) e e2)
    | app e2 e3 => app (open_ex_rec n e e2) (open_ex_rec n e e3)
    | xvar X => xvar X
    | xvar_ind n1 => xvar_ind n1
    | xabs e2 => xabs (open_ex_rec n e e2)
  end.

Definition open_eX e1 e := open_eX_rec 0 e1 e.

Definition open_ex e1 e := open_ex_rec 0 e1 e.


(* *********************************************************************** *)
(** * Size *)

Fixpoint size_e (e : exp) {struct e} : nat :=
  match e with
    | var x => 1
    | var_ind n => 1
    | abs e1 => 1 + (size_e e1)
    | app e1 e2 => 1 + (size_e e1) + (size_e e2)
    | xvar X => 1
    | xvar_ind n => 1
    | xabs e1 => 1 + (size_e e1)
  end.


(* *********************************************************************** *)
(** * Substitution *)

Fixpoint subst_eX (e : exp) (X : another_atom) (e1 : exp) {struct e1} : exp :=
  match e1 with
    | var x => var x
    | var_ind n => var_ind n
    | abs e2 => abs (subst_eX e X e2)
    | app e2 e3 => app (subst_eX e X e2) (subst_eX e X e3)
    | xvar X1 => if (X == X1) then (e) else (xvar X1)
    | xvar_ind n => xvar_ind n
    | xabs e2 => xabs (subst_eX e X e2)
  end.

Fixpoint subst_ex (e : exp) (x : atom) (e1 : exp) {struct e1} : exp :=
  match e1 with
    | var x1 => if (x == x1) then (e) else (var x1)
    | var_ind n => var_ind n
    | abs e2 => abs (subst_ex e x e2)
    | app e2 e3 => app (subst_ex e x e2) (subst_ex e x e3)
    | xvar X => xvar X
    | xvar_ind n => xvar_ind n
    | xabs e2 => xabs (subst_ex e x e2)
  end.


(* *********************************************************************** *)
(** * Degree *)

(** These define only an upper bound, not a strict upper bound. *)

Inductive degree_eX : nat -> exp -> Set :=
  | degree_eX_var : forall n x,
    degree_eX n (var x)
  | degree_eX_var_ind : forall n n1,
    degree_eX n (var_ind n1)
  | degree_eX_abs : forall n e,
    degree_eX n e ->
    degree_eX n (abs e)
  | degree_eX_app : forall n e e1,
    degree_eX n e ->
    degree_eX n e1 ->
    degree_eX n (app e e1)
  | degree_eX_xvar : forall n X,
    degree_eX n (xvar X)
  | degree_eX_xvar_ind : forall n n1,
    lt n1 n ->
    degree_eX n (xvar_ind n1)
  | degree_eX_xabs : forall n e,
    degree_eX (S n) e ->
    degree_eX n (xabs e).

Inductive degree_ex : nat -> exp -> Set :=
  | degree_ex_var : forall n x,
    degree_ex n (var x)
  | degree_ex_var_ind : forall n n1,
    lt n1 n ->
    degree_ex n (var_ind n1)
  | degree_ex_abs : forall n e,
    degree_ex (S n) e ->
    degree_ex n (abs e)
  | degree_ex_app : forall n e e1,
    degree_ex n e ->
    degree_ex n e1 ->
    degree_ex n (app e e1)
  | degree_ex_xvar : forall n X,
    degree_ex n (xvar X)
  | degree_ex_xvar_ind : forall n n1,
    degree_ex n (xvar_ind n1)
  | degree_ex_xabs : forall n e,
    degree_ex n e ->
    degree_ex n (xabs e).

Scheme degree_eX_ind' := Induction for degree_eX Sort Prop.

Definition degree_eX_mutind :=
  fun H H1 H2 H3 H4 H5 H6 H7 =>
  degree_eX_ind' H H1 H2 H3 H4 H5 H6 H7.

Scheme degree_ex_ind' := Induction for degree_ex Sort Prop.

Definition degree_ex_mutind :=
  fun H H1 H2 H3 H4 H5 H6 H7 =>
  degree_ex_ind' H H1 H2 H3 H4 H5 H6 H7.

Scheme degree_eX_rec' := Induction for degree_eX Sort Set.

Definition degree_eX_mutrec :=
  fun H H1 H2 H3 H4 H5 H6 H7 =>
  degree_eX_rec' H H1 H2 H3 H4 H5 H6 H7.

Scheme degree_ex_rec' := Induction for degree_ex Sort Set.

Definition degree_ex_mutrec :=
  fun H H1 H2 H3 H4 H5 H6 H7 =>
  degree_ex_rec' H H1 H2 H3 H4 H5 H6 H7.

Hint Constructors degree_eX : core lngen.

Hint Constructors degree_ex : core lngen.


(* *********************************************************************** *)
(** * Local closure *)

Inductive lc_e : exp -> Set :=
  | lc_e_var : forall x,
    lc_e (var x)
  | lc_e_abs : forall e,
    (forall x : atom, lc_e (open_ex (var x) e)) ->
    lc_e (abs e)
  | lc_e_app : forall e e1,
    lc_e e ->
    lc_e e1 ->
    lc_e (app e e1)
  | lc_e_xvar : forall X,
    lc_e (xvar X)
  | lc_e_xabs : forall e,
    (forall X : another_atom, lc_e (open_eX (xvar X) e)) ->
    lc_e (xabs e).

Scheme lc_e_ind' := Induction for lc_e Sort Prop.

Definition lc_e_mutind :=
  fun H H1 H2 H3 H4 H5 =>
  lc_e_ind' H H1 H2 H3 H4 H5.

Scheme lc_e_rec' := Induction for lc_e Sort Set.

Definition lc_e_mutrec :=
  fun H H1 H2 H3 H4 H5 =>
  lc_e_rec' H H1 H2 H3 H4 H5.

Hint Constructors lc_e : core lngen.


(* *********************************************************************** *)
(** * Tactic support *)

(** Tactic definitions. *)

Ltac gather_strings :=
  let S := ltac_map (fun x : string => singleton x) in
  let S1 := ltac_map (fun x : fset string => x) in
  let S2 := ltac_map (fun x : exp => fv_eX x) in
  let S3 := ltac_map (fun x : exp => fv_ex x) in
  simplify_list_of_atom_sets (S ++ S1 ++ S2 ++ S3).

Ltac pick_fresh_another_atom x :=
  let L := gather_strings in
  (pick fresh x for L).

Tactic Notation "pick" "fresh" "another_atom" ident(x) :=
  let L := gather_strings in
  (pick fresh x for L).

Ltac pick_fresh_atom x :=
  let L := gather_strings in
  (pick fresh x for L).

Tactic Notation "pick" "fresh" "atom" ident(x) :=
  let L := gather_strings in
  (pick fresh x for L).

(** Additional hint declarations. *)

Hint Resolve @plus_le_compat : lngen. (* from Coq.Arith.Plus *)

(** Redefine [default_auto] and [default_autorewrite]. *)

Ltac default_auto ::= auto with arith set brute_force lngen.
Ltac default_autorewrite ::= autorewrite with lngen in *.


(* *********************************************************************** *)
(** * Theorems about [size] *)

(* begin hide *)

Lemma size_e_min_mutual :
(forall e, 1 <= size_e e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma size_e_min :
forall e, 1 <= size_e e.
Proof.
pose proof size_e_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_min : lngen.

(* begin hide *)

Lemma size_e_close_eX_rec_mutual :
(forall e X n,
  size_e (close_eX_rec n X e) = size_e e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_e_close_eX_rec :
forall e X n,
  size_e (close_eX_rec n X e) = size_e e.
Proof.
pose proof size_e_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_close_eX_rec : lngen.
Hint Rewrite size_e_close_eX_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_e_close_ex_rec_mutual :
(forall e x n,
  size_e (close_ex_rec n x e) = size_e e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_e_close_ex_rec :
forall e x n,
  size_e (close_ex_rec n x e) = size_e e.
Proof.
pose proof size_e_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_close_ex_rec : lngen.
Hint Rewrite size_e_close_ex_rec using solve [auto] : lngen.

(* end hide *)

Lemma size_e_close_eX :
forall e X,
  size_e (close_eX X e) = size_e e.
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve size_e_close_eX : lngen.
Hint Rewrite size_e_close_eX using solve [auto] : lngen.

Lemma size_e_close_ex :
forall e x,
  size_e (close_ex x e) = size_e e.
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve size_e_close_ex : lngen.
Hint Rewrite size_e_close_ex using solve [auto] : lngen.

(* begin hide *)

Lemma size_e_open_eX_rec_mutual :
(forall e e1 n,
  size_e e <= size_e (open_eX_rec n e1 e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_e_open_eX_rec :
forall e e1 n,
  size_e e <= size_e (open_eX_rec n e1 e).
Proof.
pose proof size_e_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_open_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma size_e_open_ex_rec_mutual :
(forall e e1 n,
  size_e e <= size_e (open_ex_rec n e1 e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_e_open_ex_rec :
forall e e1 n,
  size_e e <= size_e (open_ex_rec n e1 e).
Proof.
pose proof size_e_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_open_ex_rec : lngen.

(* end hide *)

Lemma size_e_open_eX :
forall e e1,
  size_e e <= size_e (open_eX e1 e).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve size_e_open_eX : lngen.

Lemma size_e_open_ex :
forall e e1,
  size_e e <= size_e (open_ex e1 e).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve size_e_open_ex : lngen.

(* begin hide *)

Lemma size_e_open_eX_rec_var_mutual :
(forall e X n,
  size_e (open_eX_rec n (xvar X) e) = size_e e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_e_open_eX_rec_var :
forall e X n,
  size_e (open_eX_rec n (xvar X) e) = size_e e.
Proof.
pose proof size_e_open_eX_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_open_eX_rec_var : lngen.
Hint Rewrite size_e_open_eX_rec_var using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_e_open_ex_rec_var_mutual :
(forall e x n,
  size_e (open_ex_rec n (var x) e) = size_e e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_e_open_ex_rec_var :
forall e x n,
  size_e (open_ex_rec n (var x) e) = size_e e.
Proof.
pose proof size_e_open_ex_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_open_ex_rec_var : lngen.
Hint Rewrite size_e_open_ex_rec_var using solve [auto] : lngen.

(* end hide *)

Lemma size_e_open_eX_var :
forall e X,
  size_e (open_eX (xvar X) e) = size_e e.
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve size_e_open_eX_var : lngen.
Hint Rewrite size_e_open_eX_var using solve [auto] : lngen.

Lemma size_e_open_ex_var :
forall e x,
  size_e (open_ex (var x) e) = size_e e.
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve size_e_open_ex_var : lngen.
Hint Rewrite size_e_open_ex_var using solve [auto] : lngen.


(* *********************************************************************** *)
(** * Theorems about [degree] *)

(* begin hide *)

Lemma degree_eX_S_mutual :
(forall n e,
  degree_eX n e ->
  degree_eX (S n) e).
Proof.
apply_mutual_ind degree_eX_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma degree_eX_S :
forall n e,
  degree_eX n e ->
  degree_eX (S n) e.
Proof.
pose proof degree_eX_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_eX_S : lngen.

(* begin hide *)

Lemma degree_ex_S_mutual :
(forall n e,
  degree_ex n e ->
  degree_ex (S n) e).
Proof.
apply_mutual_ind degree_ex_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma degree_ex_S :
forall n e,
  degree_ex n e ->
  degree_ex (S n) e.
Proof.
pose proof degree_ex_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ex_S : lngen.

Lemma degree_eX_O :
forall n e,
  degree_eX O e ->
  degree_eX n e.
Proof.
induction n; default_simp.
Qed.

Hint Resolve degree_eX_O : lngen.

Lemma degree_ex_O :
forall n e,
  degree_ex O e ->
  degree_ex n e.
Proof.
induction n; default_simp.
Qed.

Hint Resolve degree_ex_O : lngen.

(* begin hide *)

Lemma degree_eX_close_eX_rec_mutual :
(forall e X n,
  degree_eX n e ->
  degree_eX (S n) (close_eX_rec n X e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_eX_rec :
forall e X n,
  degree_eX n e ->
  degree_eX (S n) (close_eX_rec n X e).
Proof.
pose proof degree_eX_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_eX_close_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_ex_rec_mutual :
(forall e x n n1,
  degree_eX n e ->
  degree_eX n (close_ex_rec n1 x e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_ex_rec :
forall e x n n1,
  degree_eX n e ->
  degree_eX n (close_ex_rec n1 x e).
Proof.
pose proof degree_eX_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_eX_close_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_eX_rec_mutual :
(forall e X n n1,
  degree_ex n e ->
  degree_ex n (close_eX_rec n1 X e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_eX_rec :
forall e X n n1,
  degree_ex n e ->
  degree_ex n (close_eX_rec n1 X e).
Proof.
pose proof degree_ex_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ex_close_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_ex_rec_mutual :
(forall e x n,
  degree_ex n e ->
  degree_ex (S n) (close_ex_rec n x e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_ex_rec :
forall e x n,
  degree_ex n e ->
  degree_ex (S n) (close_ex_rec n x e).
Proof.
pose proof degree_ex_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ex_close_ex_rec : lngen.

(* end hide *)

Lemma degree_eX_close_eX :
forall e X,
  degree_eX 0 e ->
  degree_eX 1 (close_eX X e).
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve degree_eX_close_eX : lngen.

Lemma degree_eX_close_ex :
forall e x n,
  degree_eX n e ->
  degree_eX n (close_ex x e).
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve degree_eX_close_ex : lngen.

Lemma degree_ex_close_eX :
forall e X n,
  degree_ex n e ->
  degree_ex n (close_eX X e).
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve degree_ex_close_eX : lngen.

Lemma degree_ex_close_ex :
forall e x,
  degree_ex 0 e ->
  degree_ex 1 (close_ex x e).
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve degree_ex_close_ex : lngen.

(* begin hide *)

Lemma degree_eX_close_eX_rec_inv_mutual :
(forall e X n,
  degree_eX (S n) (close_eX_rec n X e) ->
  degree_eX n e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_eX_rec_inv :
forall e X n,
  degree_eX (S n) (close_eX_rec n X e) ->
  degree_eX n e.
Proof.
pose proof degree_eX_close_eX_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_eX_close_eX_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_ex_rec_inv_mutual :
(forall e x n n1,
  degree_eX n (close_ex_rec n1 x e) ->
  degree_eX n e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_ex_rec_inv :
forall e x n n1,
  degree_eX n (close_ex_rec n1 x e) ->
  degree_eX n e.
Proof.
pose proof degree_eX_close_ex_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_eX_close_ex_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_eX_rec_inv_mutual :
(forall e X n n1,
  degree_ex n (close_eX_rec n1 X e) ->
  degree_ex n e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_eX_rec_inv :
forall e X n n1,
  degree_ex n (close_eX_rec n1 X e) ->
  degree_ex n e.
Proof.
pose proof degree_ex_close_eX_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_ex_close_eX_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_ex_rec_inv_mutual :
(forall e x n,
  degree_ex (S n) (close_ex_rec n x e) ->
  degree_ex n e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_ex_rec_inv :
forall e x n,
  degree_ex (S n) (close_ex_rec n x e) ->
  degree_ex n e.
Proof.
pose proof degree_ex_close_ex_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_ex_close_ex_rec_inv : lngen.

(* end hide *)

Lemma degree_eX_close_eX_inv :
forall e X,
  degree_eX 1 (close_eX X e) ->
  degree_eX 0 e.
Proof.
unfold close_eX; eauto 2 with lngen.
Qed.

Hint Immediate degree_eX_close_eX_inv : lngen.

Lemma degree_eX_close_ex_inv :
forall e x n,
  degree_eX n (close_ex x e) ->
  degree_eX n e.
Proof.
unfold close_ex; eauto 2 with lngen.
Qed.

Hint Immediate degree_eX_close_ex_inv : lngen.

Lemma degree_ex_close_eX_inv :
forall e X n,
  degree_ex n (close_eX X e) ->
  degree_ex n e.
Proof.
unfold close_eX; eauto 2 with lngen.
Qed.

Hint Immediate degree_ex_close_eX_inv : lngen.

Lemma degree_ex_close_ex_inv :
forall e x,
  degree_ex 1 (close_ex x e) ->
  degree_ex 0 e.
Proof.
unfold close_ex; eauto 2 with lngen.
Qed.

Hint Immediate degree_ex_close_ex_inv : lngen.

(* begin hide *)

Lemma degree_eX_open_eX_rec_mutual :
(forall e e1 n,
  degree_eX (S n) e ->
  degree_eX n e1 ->
  degree_eX n (open_eX_rec n e1 e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_eX_rec :
forall e e1 n,
  degree_eX (S n) e ->
  degree_eX n e1 ->
  degree_eX n (open_eX_rec n e1 e).
Proof.
pose proof degree_eX_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_eX_open_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_ex_rec_mutual :
(forall e e1 n n1,
  degree_eX n e ->
  degree_eX n e1 ->
  degree_eX n (open_ex_rec n1 e1 e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_ex_rec :
forall e e1 n n1,
  degree_eX n e ->
  degree_eX n e1 ->
  degree_eX n (open_ex_rec n1 e1 e).
Proof.
pose proof degree_eX_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_eX_open_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_eX_rec_mutual :
(forall e e1 n n1,
  degree_ex n e ->
  degree_ex n e1 ->
  degree_ex n (open_eX_rec n1 e1 e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_eX_rec :
forall e e1 n n1,
  degree_ex n e ->
  degree_ex n e1 ->
  degree_ex n (open_eX_rec n1 e1 e).
Proof.
pose proof degree_ex_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ex_open_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_ex_rec_mutual :
(forall e e1 n,
  degree_ex (S n) e ->
  degree_ex n e1 ->
  degree_ex n (open_ex_rec n e1 e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_ex_rec :
forall e e1 n,
  degree_ex (S n) e ->
  degree_ex n e1 ->
  degree_ex n (open_ex_rec n e1 e).
Proof.
pose proof degree_ex_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ex_open_ex_rec : lngen.

(* end hide *)

Lemma degree_eX_open_eX :
forall e e1,
  degree_eX 1 e ->
  degree_eX 0 e1 ->
  degree_eX 0 (open_eX e1 e).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve degree_eX_open_eX : lngen.

Lemma degree_eX_open_ex :
forall e e1 n,
  degree_eX n e ->
  degree_eX n e1 ->
  degree_eX n (open_ex e1 e).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve degree_eX_open_ex : lngen.

Lemma degree_ex_open_eX :
forall e e1 n,
  degree_ex n e ->
  degree_ex n e1 ->
  degree_ex n (open_eX e1 e).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve degree_ex_open_eX : lngen.

Lemma degree_ex_open_ex :
forall e e1,
  degree_ex 1 e ->
  degree_ex 0 e1 ->
  degree_ex 0 (open_ex e1 e).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve degree_ex_open_ex : lngen.

(* begin hide *)

Lemma degree_eX_open_eX_rec_inv_mutual :
(forall e e1 n,
  degree_eX n (open_eX_rec n e1 e) ->
  degree_eX (S n) e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_eX_rec_inv :
forall e e1 n,
  degree_eX n (open_eX_rec n e1 e) ->
  degree_eX (S n) e.
Proof.
pose proof degree_eX_open_eX_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_eX_open_eX_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_ex_rec_inv_mutual :
(forall e e1 n n1,
  degree_eX n (open_ex_rec n1 e1 e) ->
  degree_eX n e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_ex_rec_inv :
forall e e1 n n1,
  degree_eX n (open_ex_rec n1 e1 e) ->
  degree_eX n e.
Proof.
pose proof degree_eX_open_ex_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_eX_open_ex_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_eX_rec_inv_mutual :
(forall e e1 n n1,
  degree_ex n (open_eX_rec n1 e1 e) ->
  degree_ex n e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_eX_rec_inv :
forall e e1 n n1,
  degree_ex n (open_eX_rec n1 e1 e) ->
  degree_ex n e.
Proof.
pose proof degree_ex_open_eX_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_ex_open_eX_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_ex_rec_inv_mutual :
(forall e e1 n,
  degree_ex n (open_ex_rec n e1 e) ->
  degree_ex (S n) e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_ex_rec_inv :
forall e e1 n,
  degree_ex n (open_ex_rec n e1 e) ->
  degree_ex (S n) e.
Proof.
pose proof degree_ex_open_ex_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_ex_open_ex_rec_inv : lngen.

(* end hide *)

Lemma degree_eX_open_eX_inv :
forall e e1,
  degree_eX 0 (open_eX e1 e) ->
  degree_eX 1 e.
Proof.
unfold open_eX; eauto 2 with lngen.
Qed.

Hint Resolve degree_eX_open_eX_inv : lngen.

Lemma degree_eX_open_ex_inv :
forall e e1 n,
  degree_eX n (open_ex e1 e) ->
  degree_eX n e.
Proof.
unfold open_ex; eauto 2 with lngen.
Qed.

Hint Resolve degree_eX_open_ex_inv : lngen.

Lemma degree_ex_open_eX_inv :
forall e e1 n,
  degree_ex n (open_eX e1 e) ->
  degree_ex n e.
Proof.
unfold open_eX; eauto 2 with lngen.
Qed.

Hint Resolve degree_ex_open_eX_inv : lngen.

Lemma degree_ex_open_ex_inv :
forall e e1,
  degree_ex 0 (open_ex e1 e) ->
  degree_ex 1 e.
Proof.
unfold open_ex; eauto 2 with lngen.
Qed.

Hint Resolve degree_ex_open_ex_inv : lngen.


(* *********************************************************************** *)
(** * Theorems about [open] and [close] *)

(* begin hide *)

Lemma close_eX_rec_inj_mutual :
(forall e1 e X n,
  close_eX_rec n X e1 = close_eX_rec n X e ->
  e1 = e).
Proof.
apply_mutual_ind exp_mutind;
intros; (destruct e);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_eX_rec_inj :
forall e1 e X n,
  close_eX_rec n X e1 = close_eX_rec n X e ->
  e1 = e.
Proof.
pose proof close_eX_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_eX_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma close_ex_rec_inj_mutual :
(forall e1 e x n,
  close_ex_rec n x e1 = close_ex_rec n x e ->
  e1 = e).
Proof.
apply_mutual_ind exp_mutind;
intros; (destruct e);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_ex_rec_inj :
forall e1 e x n,
  close_ex_rec n x e1 = close_ex_rec n x e ->
  e1 = e.
Proof.
pose proof close_ex_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_ex_rec_inj : lngen.

(* end hide *)

Lemma close_eX_inj :
forall e1 e X,
  close_eX X e1 = close_eX X e ->
  e1 = e.
Proof.
unfold close_eX; eauto 2 with lngen.
Qed.

Hint Immediate close_eX_inj : lngen.

Lemma close_ex_inj :
forall e1 e x,
  close_ex x e1 = close_ex x e ->
  e1 = e.
Proof.
unfold close_ex; eauto 2 with lngen.
Qed.

Hint Immediate close_ex_inj : lngen.

(* begin hide *)

Lemma close_eX_rec_open_eX_rec_mutual :
(forall e X n,
  X `notin` fv_eX e ->
  close_eX_rec n X (open_eX_rec n (xvar X) e) = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_eX_rec_open_eX_rec :
forall e X n,
  X `notin` fv_eX e ->
  close_eX_rec n X (open_eX_rec n (xvar X) e) = e.
Proof.
pose proof close_eX_rec_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_eX_rec_open_eX_rec : lngen.
Hint Rewrite close_eX_rec_open_eX_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_ex_rec_open_ex_rec_mutual :
(forall e x n,
  x `notin` fv_ex e ->
  close_ex_rec n x (open_ex_rec n (var x) e) = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_ex_rec_open_ex_rec :
forall e x n,
  x `notin` fv_ex e ->
  close_ex_rec n x (open_ex_rec n (var x) e) = e.
Proof.
pose proof close_ex_rec_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_ex_rec_open_ex_rec : lngen.
Hint Rewrite close_ex_rec_open_ex_rec using solve [auto] : lngen.

(* end hide *)

Lemma close_eX_open_eX :
forall e X,
  X `notin` fv_eX e ->
  close_eX X (open_eX (xvar X) e) = e.
Proof.
unfold close_eX; unfold open_eX; default_simp.
Qed.

Hint Resolve close_eX_open_eX : lngen.
Hint Rewrite close_eX_open_eX using solve [auto] : lngen.

Lemma close_ex_open_ex :
forall e x,
  x `notin` fv_ex e ->
  close_ex x (open_ex (var x) e) = e.
Proof.
unfold close_ex; unfold open_ex; default_simp.
Qed.

Hint Resolve close_ex_open_ex : lngen.
Hint Rewrite close_ex_open_ex using solve [auto] : lngen.

(* begin hide *)

Lemma open_eX_rec_close_eX_rec_mutual :
(forall e X n,
  open_eX_rec n (xvar X) (close_eX_rec n X e) = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_eX_rec_close_eX_rec :
forall e X n,
  open_eX_rec n (xvar X) (close_eX_rec n X e) = e.
Proof.
pose proof open_eX_rec_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_eX_rec_close_eX_rec : lngen.
Hint Rewrite open_eX_rec_close_eX_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_ex_rec_close_ex_rec_mutual :
(forall e x n,
  open_ex_rec n (var x) (close_ex_rec n x e) = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_ex_rec_close_ex_rec :
forall e x n,
  open_ex_rec n (var x) (close_ex_rec n x e) = e.
Proof.
pose proof open_ex_rec_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_ex_rec_close_ex_rec : lngen.
Hint Rewrite open_ex_rec_close_ex_rec using solve [auto] : lngen.

(* end hide *)

Lemma open_eX_close_eX :
forall e X,
  open_eX (xvar X) (close_eX X e) = e.
Proof.
unfold close_eX; unfold open_eX; default_simp.
Qed.

Hint Resolve open_eX_close_eX : lngen.
Hint Rewrite open_eX_close_eX using solve [auto] : lngen.

Lemma open_ex_close_ex :
forall e x,
  open_ex (var x) (close_ex x e) = e.
Proof.
unfold close_ex; unfold open_ex; default_simp.
Qed.

Hint Resolve open_ex_close_ex : lngen.
Hint Rewrite open_ex_close_ex using solve [auto] : lngen.

(* begin hide *)

Lemma open_eX_rec_inj_mutual :
(forall e1 e X n,
  X `notin` fv_eX e1 ->
  X `notin` fv_eX e ->
  open_eX_rec n (xvar X) e1 = open_eX_rec n (xvar X) e ->
  e1 = e).
Proof.
apply_mutual_ind exp_mutind;
intros; (destruct e);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_eX_rec_inj :
forall e1 e X n,
  X `notin` fv_eX e1 ->
  X `notin` fv_eX e ->
  open_eX_rec n (xvar X) e1 = open_eX_rec n (xvar X) e ->
  e1 = e.
Proof.
pose proof open_eX_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_eX_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma open_ex_rec_inj_mutual :
(forall e1 e x n,
  x `notin` fv_ex e1 ->
  x `notin` fv_ex e ->
  open_ex_rec n (var x) e1 = open_ex_rec n (var x) e ->
  e1 = e).
Proof.
apply_mutual_ind exp_mutind;
intros; (destruct e);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_ex_rec_inj :
forall e1 e x n,
  x `notin` fv_ex e1 ->
  x `notin` fv_ex e ->
  open_ex_rec n (var x) e1 = open_ex_rec n (var x) e ->
  e1 = e.
Proof.
pose proof open_ex_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_ex_rec_inj : lngen.

(* end hide *)

Lemma open_eX_inj :
forall e1 e X,
  X `notin` fv_eX e1 ->
  X `notin` fv_eX e ->
  open_eX (xvar X) e1 = open_eX (xvar X) e ->
  e1 = e.
Proof.
unfold open_eX; eauto 2 with lngen.
Qed.

Hint Immediate open_eX_inj : lngen.

Lemma open_ex_inj :
forall e1 e x,
  x `notin` fv_ex e1 ->
  x `notin` fv_ex e ->
  open_ex (var x) e1 = open_ex (var x) e ->
  e1 = e.
Proof.
unfold open_ex; eauto 2 with lngen.
Qed.

Hint Immediate open_ex_inj : lngen.


(* *********************************************************************** *)
(** * Theorems about [lc] *)

(* begin hide *)

Lemma degree_eX_of_lc_e_mutual :
(forall e,
  lc_e e ->
  degree_eX 0 e).
Proof.
apply_mutual_ind lc_e_mutrec;
intros;
let X := fresh "X" in pick_fresh_another_atom X;
let x := fresh "x" in pick_fresh_atom x;
repeat (match goal with
          | H : _, H1 : _ |- _ => specialize H with H1
        end);
default_simp; eauto with lngen.
Qed.

(* end hide *)

Lemma degree_eX_of_lc_e :
forall e,
  lc_e e ->
  degree_eX 0 e.
Proof.
pose proof degree_eX_of_lc_e_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_eX_of_lc_e : lngen.

(* begin hide *)

Lemma degree_ex_of_lc_e_mutual :
(forall e,
  lc_e e ->
  degree_ex 0 e).
Proof.
apply_mutual_ind lc_e_mutrec;
intros;
let X := fresh "X" in pick_fresh_another_atom X;
let x := fresh "x" in pick_fresh_atom x;
repeat (match goal with
          | H : _, H1 : _ |- _ => specialize H with H1
        end);
default_simp; eauto with lngen.
Qed.

(* end hide *)

Lemma degree_ex_of_lc_e :
forall e,
  lc_e e ->
  degree_ex 0 e.
Proof.
pose proof degree_ex_of_lc_e_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ex_of_lc_e : lngen.

(* begin hide *)

Lemma lc_e_of_degree_size_mutual :
forall i,
(forall e,
  size_e e = i ->
  degree_eX 0 e ->
  degree_ex 0 e ->
  lc_e e).
Proof.
intros i; pattern i; apply lt_wf_rec;
clear i; intros i H;
apply_mutual_ind exp_mutrec;
default_simp;
(* non-trivial cases *)
constructor; default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_e_of_degree :
forall e,
  degree_eX 0 e ->
  degree_ex 0 e ->
  lc_e e.
Proof.
intros e; intros;
pose proof (lc_e_of_degree_size_mutual (size_e e));
intuition eauto.
Qed.

Hint Resolve lc_e_of_degree : lngen.

Lemma lc_e_abs_exists :
forall x e,
  lc_e (open_ex (var x) e) ->
  lc_e (abs e).
Proof.
default_simp; eauto 10 with lngen.
Qed.

Hint Resolve lc_e_abs_exists : lngen.

Lemma lc_e_xabs_exists :
forall X e,
  lc_e (open_eX (xvar X) e) ->
  lc_e (xabs e).
Proof.
default_simp; eauto 10 with lngen.
Qed.

Hint Resolve lc_e_xabs_exists : lngen.

(* begin hide *)

Lemma lc_e_unique_mutual :
(forall e (proof proof1 : lc_e e), proof = proof1).
Proof.
apply_mutual_ind lc_e_mutind;
intros; dependent destruction proof1; default_simp; auto using axiom_fun_ext_eq.
Qed.

(* end hide *)

Lemma lc_e_unique :
forall e (proof proof1 : lc_e e), proof = proof1.
Proof.
pose proof lc_e_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_e_unique : lngen.


(* *********************************************************************** *)
(** * More theorems about [open] and [close] *)

(* begin hide *)

Lemma close_eX_rec_degree_eX_mutual :
(forall e X n,
  degree_eX n e ->
  X `notin` fv_eX e ->
  close_eX_rec n X e = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_eX_rec_degree_eX :
forall e X n,
  degree_eX n e ->
  X `notin` fv_eX e ->
  close_eX_rec n X e = e.
Proof.
pose proof close_eX_rec_degree_eX_mutual as H; intuition eauto.
Qed.

Hint Resolve close_eX_rec_degree_eX : lngen.
Hint Rewrite close_eX_rec_degree_eX using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_ex_rec_degree_ex_mutual :
(forall e x n,
  degree_ex n e ->
  x `notin` fv_ex e ->
  close_ex_rec n x e = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_ex_rec_degree_ex :
forall e x n,
  degree_ex n e ->
  x `notin` fv_ex e ->
  close_ex_rec n x e = e.
Proof.
pose proof close_ex_rec_degree_ex_mutual as H; intuition eauto.
Qed.

Hint Resolve close_ex_rec_degree_ex : lngen.
Hint Rewrite close_ex_rec_degree_ex using solve [auto] : lngen.

(* end hide *)

Lemma close_eX_lc_e :
forall e X,
  lc_e e ->
  X `notin` fv_eX e ->
  close_eX X e = e.
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve close_eX_lc_e : lngen.
Hint Rewrite close_eX_lc_e using solve [auto] : lngen.

Lemma close_ex_lc_e :
forall e x,
  lc_e e ->
  x `notin` fv_ex e ->
  close_ex x e = e.
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve close_ex_lc_e : lngen.
Hint Rewrite close_ex_lc_e using solve [auto] : lngen.

(* begin hide *)

Lemma open_eX_rec_degree_eX_mutual :
(forall e1 e n,
  degree_eX n e1 ->
  open_eX_rec n e e1 = e1).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_eX_rec_degree_eX :
forall e1 e n,
  degree_eX n e1 ->
  open_eX_rec n e e1 = e1.
Proof.
pose proof open_eX_rec_degree_eX_mutual as H; intuition eauto.
Qed.

Hint Resolve open_eX_rec_degree_eX : lngen.
Hint Rewrite open_eX_rec_degree_eX using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_ex_rec_degree_ex_mutual :
(forall e1 e n,
  degree_ex n e1 ->
  open_ex_rec n e e1 = e1).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_ex_rec_degree_ex :
forall e1 e n,
  degree_ex n e1 ->
  open_ex_rec n e e1 = e1.
Proof.
pose proof open_ex_rec_degree_ex_mutual as H; intuition eauto.
Qed.

Hint Resolve open_ex_rec_degree_ex : lngen.
Hint Rewrite open_ex_rec_degree_ex using solve [auto] : lngen.

(* end hide *)

Lemma open_eX_lc_e :
forall e1 e,
  lc_e e1 ->
  open_eX e e1 = e1.
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve open_eX_lc_e : lngen.
Hint Rewrite open_eX_lc_e using solve [auto] : lngen.

Lemma open_ex_lc_e :
forall e1 e,
  lc_e e1 ->
  open_ex e e1 = e1.
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve open_ex_lc_e : lngen.
Hint Rewrite open_ex_lc_e using solve [auto] : lngen.


(* *********************************************************************** *)
(** * Theorems about [subst] *)

(* begin hide *)

Lemma subst_eX_close_eX_rec_mutual :
(forall e1 e X X1 n,
  degree_eX n e ->
  X <> X1 ->
  X1 `notin` fv_eX e ->
  subst_eX e X (close_eX_rec n X1 e1) = close_eX_rec n X1 (subst_eX e X e1)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_close_eX_rec :
forall e1 e X X1 n,
  degree_eX n e ->
  X <> X1 ->
  X1 `notin` fv_eX e ->
  subst_eX e X (close_eX_rec n X1 e1) = close_eX_rec n X1 (subst_eX e X e1).
Proof.
pose proof subst_eX_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_close_eX_rec : lngen.

(* begin hide *)

Lemma subst_eX_close_ex_rec_mutual :
(forall e1 e x X n,
  degree_ex n e ->
  X `notin` fv_ex e ->
  subst_eX e x (close_ex_rec n X e1) = close_ex_rec n X (subst_eX e x e1)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_close_ex_rec :
forall e1 e x X n,
  degree_ex n e ->
  X `notin` fv_ex e ->
  subst_eX e x (close_ex_rec n X e1) = close_ex_rec n X (subst_eX e x e1).
Proof.
pose proof subst_eX_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_close_ex_rec : lngen.

(* begin hide *)

Lemma subst_ex_close_eX_rec_mutual :
(forall e1 e X x n,
  degree_eX n e ->
  x `notin` fv_eX e ->
  subst_ex e X (close_eX_rec n x e1) = close_eX_rec n x (subst_ex e X e1)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_close_eX_rec :
forall e1 e X x n,
  degree_eX n e ->
  x `notin` fv_eX e ->
  subst_ex e X (close_eX_rec n x e1) = close_eX_rec n x (subst_ex e X e1).
Proof.
pose proof subst_ex_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_close_eX_rec : lngen.

(* begin hide *)

Lemma subst_ex_close_ex_rec_mutual :
(forall e1 e x x1 n,
  degree_ex n e ->
  x <> x1 ->
  x1 `notin` fv_ex e ->
  subst_ex e x (close_ex_rec n x1 e1) = close_ex_rec n x1 (subst_ex e x e1)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_close_ex_rec :
forall e1 e x x1 n,
  degree_ex n e ->
  x <> x1 ->
  x1 `notin` fv_ex e ->
  subst_ex e x (close_ex_rec n x1 e1) = close_ex_rec n x1 (subst_ex e x e1).
Proof.
pose proof subst_ex_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_close_ex_rec : lngen.

Lemma subst_eX_close_eX :
forall e1 e X X1,
  lc_e e ->  X <> X1 ->
  X1 `notin` fv_eX e ->
  subst_eX e X (close_eX X1 e1) = close_eX X1 (subst_eX e X e1).
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve subst_eX_close_eX : lngen.

Lemma subst_eX_close_ex :
forall e1 e x X,
  lc_e e ->  X `notin` fv_ex e ->
  subst_eX e x (close_ex X e1) = close_ex X (subst_eX e x e1).
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve subst_eX_close_ex : lngen.

Lemma subst_ex_close_eX :
forall e1 e X x,
  lc_e e ->  x `notin` fv_eX e ->
  subst_ex e X (close_eX x e1) = close_eX x (subst_ex e X e1).
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve subst_ex_close_eX : lngen.

Lemma subst_ex_close_ex :
forall e1 e x x1,
  lc_e e ->  x <> x1 ->
  x1 `notin` fv_ex e ->
  subst_ex e x (close_ex x1 e1) = close_ex x1 (subst_ex e x e1).
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve subst_ex_close_ex : lngen.

(* begin hide *)

Lemma subst_eX_degree_eX_mutual :
(forall e e1 X n,
  degree_eX n e ->
  degree_eX n e1 ->
  degree_eX n (subst_eX e1 X e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_degree_eX :
forall e e1 X n,
  degree_eX n e ->
  degree_eX n e1 ->
  degree_eX n (subst_eX e1 X e).
Proof.
pose proof subst_eX_degree_eX_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_degree_eX : lngen.

(* begin hide *)

Lemma subst_eX_degree_ex_mutual :
(forall e e1 X n,
  degree_ex n e ->
  degree_ex n e1 ->
  degree_ex n (subst_eX e1 X e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_degree_ex :
forall e e1 X n,
  degree_ex n e ->
  degree_ex n e1 ->
  degree_ex n (subst_eX e1 X e).
Proof.
pose proof subst_eX_degree_ex_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_degree_ex : lngen.

(* begin hide *)

Lemma subst_ex_degree_eX_mutual :
(forall e e1 x n,
  degree_eX n e ->
  degree_eX n e1 ->
  degree_eX n (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_degree_eX :
forall e e1 x n,
  degree_eX n e ->
  degree_eX n e1 ->
  degree_eX n (subst_ex e1 x e).
Proof.
pose proof subst_ex_degree_eX_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_degree_eX : lngen.

(* begin hide *)

Lemma subst_ex_degree_ex_mutual :
(forall e e1 x n,
  degree_ex n e ->
  degree_ex n e1 ->
  degree_ex n (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_degree_ex :
forall e e1 x n,
  degree_ex n e ->
  degree_ex n e1 ->
  degree_ex n (subst_ex e1 x e).
Proof.
pose proof subst_ex_degree_ex_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_degree_ex : lngen.

(* begin hide *)

Lemma subst_eX_fresh_mutual :
(forall e1 e X,
  X `notin` fv_eX e1 ->
  subst_eX e X e1 = e1).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_fresh :
forall e1 e X,
  X `notin` fv_eX e1 ->
  subst_eX e X e1 = e1.
Proof.
pose proof subst_eX_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_fresh : lngen.
Hint Rewrite subst_eX_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma subst_ex_fresh_mutual :
(forall e1 e x,
  x `notin` fv_ex e1 ->
  subst_ex e x e1 = e1).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_fresh :
forall e1 e x,
  x `notin` fv_ex e1 ->
  subst_ex e x e1 = e1.
Proof.
pose proof subst_ex_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_fresh : lngen.
Hint Rewrite subst_ex_fresh using solve [auto] : lngen.

Lemma subst_eX_lc_e :
forall e e1 X,
  lc_e e ->
  lc_e e1 ->
  lc_e (subst_eX e1 X e).
Proof.
default_simp.
Qed.

Hint Resolve subst_eX_lc_e : lngen.

Lemma subst_ex_lc_e :
forall e e1 x,
  lc_e e ->
  lc_e e1 ->
  lc_e (subst_ex e1 x e).
Proof.
default_simp.
Qed.

Hint Resolve subst_ex_lc_e : lngen.

(* begin hide *)

Lemma subst_eX_open_eX_rec_mutual :
(forall e2 e e1 X n,
  lc_e e ->
  subst_eX e X (open_eX_rec n e1 e2) = open_eX_rec n (subst_eX e X e1) (subst_eX e X e2)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_eX_open_eX_rec :
forall e2 e e1 X n,
  lc_e e ->
  subst_eX e X (open_eX_rec n e1 e2) = open_eX_rec n (subst_eX e X e1) (subst_eX e X e2).
Proof.
pose proof subst_eX_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_open_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_eX_open_ex_rec_mutual :
(forall e2 e e1 X n,
  lc_e e ->
  subst_eX e X (open_ex_rec n e1 e2) = open_ex_rec n (subst_eX e X e1) (subst_eX e X e2)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_eX_open_ex_rec :
forall e2 e e1 X n,
  lc_e e ->
  subst_eX e X (open_ex_rec n e1 e2) = open_ex_rec n (subst_eX e X e1) (subst_eX e X e2).
Proof.
pose proof subst_eX_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_open_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_ex_open_eX_rec_mutual :
(forall e2 e e1 x n,
  lc_e e ->
  subst_ex e x (open_eX_rec n e1 e2) = open_eX_rec n (subst_ex e x e1) (subst_ex e x e2)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_ex_open_eX_rec :
forall e2 e e1 x n,
  lc_e e ->
  subst_ex e x (open_eX_rec n e1 e2) = open_eX_rec n (subst_ex e x e1) (subst_ex e x e2).
Proof.
pose proof subst_ex_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_open_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_ex_open_ex_rec_mutual :
(forall e2 e e1 x n,
  lc_e e ->
  subst_ex e x (open_ex_rec n e1 e2) = open_ex_rec n (subst_ex e x e1) (subst_ex e x e2)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_ex_open_ex_rec :
forall e2 e e1 x n,
  lc_e e ->
  subst_ex e x (open_ex_rec n e1 e2) = open_ex_rec n (subst_ex e x e1) (subst_ex e x e2).
Proof.
pose proof subst_ex_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_open_ex_rec : lngen.

(* end hide *)

Lemma subst_eX_open_eX :
forall e2 e e1 X,
  lc_e e ->
  subst_eX e X (open_eX e1 e2) = open_eX (subst_eX e X e1) (subst_eX e X e2).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve subst_eX_open_eX : lngen.

Lemma subst_eX_open_ex :
forall e2 e e1 X,
  lc_e e ->
  subst_eX e X (open_ex e1 e2) = open_ex (subst_eX e X e1) (subst_eX e X e2).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve subst_eX_open_ex : lngen.

Lemma subst_ex_open_eX :
forall e2 e e1 x,
  lc_e e ->
  subst_ex e x (open_eX e1 e2) = open_eX (subst_ex e x e1) (subst_ex e x e2).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve subst_ex_open_eX : lngen.

Lemma subst_ex_open_ex :
forall e2 e e1 x,
  lc_e e ->
  subst_ex e x (open_ex e1 e2) = open_ex (subst_ex e x e1) (subst_ex e x e2).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve subst_ex_open_ex : lngen.

Lemma subst_eX_open_eX_var :
forall e1 e X X1,
  X <> X1 ->
  lc_e e ->
  subst_eX e X (open_eX (xvar X1) e1) = open_eX (xvar X1) (subst_eX e X e1).
Proof.
intros; rewrite subst_eX_open_eX; default_simp.
Qed.

Hint Resolve subst_eX_open_eX_var : lngen.

Lemma subst_eX_open_ex_var :
forall e1 e X x,
  lc_e e ->
  subst_eX e X (open_ex (var x) e1) = open_ex (var x) (subst_eX e X e1).
Proof.
intros; rewrite subst_eX_open_ex; default_simp.
Qed.

Hint Resolve subst_eX_open_ex_var : lngen.

Lemma subst_ex_open_eX_var :
forall e1 e x X,
  lc_e e ->
  subst_ex e x (open_eX (xvar X) e1) = open_eX (xvar X) (subst_ex e x e1).
Proof.
intros; rewrite subst_ex_open_eX; default_simp.
Qed.

Hint Resolve subst_ex_open_eX_var : lngen.

Lemma subst_ex_open_ex_var :
forall e1 e x x1,
  x <> x1 ->
  lc_e e ->
  subst_ex e x (open_ex (var x1) e1) = open_ex (var x1) (subst_ex e x e1).
Proof.
intros; rewrite subst_ex_open_ex; default_simp.
Qed.

Hint Resolve subst_ex_open_ex_var : lngen.

(* begin hide *)

Lemma subst_eX_spec_rec_mutual :
(forall e e1 X n,
  subst_eX e1 X e = open_eX_rec n e1 (close_eX_rec n X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_eX_spec_rec :
forall e e1 X n,
  subst_eX e1 X e = open_eX_rec n e1 (close_eX_rec n X e).
Proof.
pose proof subst_eX_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_spec_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_ex_spec_rec_mutual :
(forall e e1 x n,
  subst_ex e1 x e = open_ex_rec n e1 (close_ex_rec n x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_ex_spec_rec :
forall e e1 x n,
  subst_ex e1 x e = open_ex_rec n e1 (close_ex_rec n x e).
Proof.
pose proof subst_ex_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_spec_rec : lngen.

(* end hide *)

Lemma subst_eX_spec :
forall e e1 X,
  subst_eX e1 X e = open_eX e1 (close_eX X e).
Proof.
unfold close_eX; unfold open_eX; default_simp.
Qed.

Hint Resolve subst_eX_spec : lngen.

Lemma subst_ex_spec :
forall e e1 x,
  subst_ex e1 x e = open_ex e1 (close_ex x e).
Proof.
unfold close_ex; unfold open_ex; default_simp.
Qed.

Hint Resolve subst_ex_spec : lngen.

(* begin hide *)

Lemma subst_eX_subst_eX_mutual :
(forall e e1 e2 X1 X,
  X1 `notin` fv_eX e1 ->
  X1 <> X ->
  subst_eX e1 X (subst_eX e2 X1 e) = subst_eX (subst_eX e1 X e2) X1 (subst_eX e1 X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_subst_eX :
forall e e1 e2 X1 X,
  X1 `notin` fv_eX e1 ->
  X1 <> X ->
  subst_eX e1 X (subst_eX e2 X1 e) = subst_eX (subst_eX e1 X e2) X1 (subst_eX e1 X e).
Proof.
pose proof subst_eX_subst_eX_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_subst_eX : lngen.

(* begin hide *)

Lemma subst_eX_subst_ex_mutual :
(forall e e1 e2 x X,
  x `notin` fv_ex e1 ->
  subst_eX e1 X (subst_ex e2 x e) = subst_ex (subst_eX e1 X e2) x (subst_eX e1 X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_subst_ex :
forall e e1 e2 x X,
  x `notin` fv_ex e1 ->
  subst_eX e1 X (subst_ex e2 x e) = subst_ex (subst_eX e1 X e2) x (subst_eX e1 X e).
Proof.
pose proof subst_eX_subst_ex_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_subst_ex : lngen.

(* begin hide *)

Lemma subst_ex_subst_eX_mutual :
(forall e e1 e2 X x,
  X `notin` fv_eX e1 ->
  subst_ex e1 x (subst_eX e2 X e) = subst_eX (subst_ex e1 x e2) X (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_subst_eX :
forall e e1 e2 X x,
  X `notin` fv_eX e1 ->
  subst_ex e1 x (subst_eX e2 X e) = subst_eX (subst_ex e1 x e2) X (subst_ex e1 x e).
Proof.
pose proof subst_ex_subst_eX_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_subst_eX : lngen.

(* begin hide *)

Lemma subst_ex_subst_ex_mutual :
(forall e e1 e2 x1 x,
  x1 `notin` fv_ex e1 ->
  x1 <> x ->
  subst_ex e1 x (subst_ex e2 x1 e) = subst_ex (subst_ex e1 x e2) x1 (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_subst_ex :
forall e e1 e2 x1 x,
  x1 `notin` fv_ex e1 ->
  x1 <> x ->
  subst_ex e1 x (subst_ex e2 x1 e) = subst_ex (subst_ex e1 x e2) x1 (subst_ex e1 x e).
Proof.
pose proof subst_ex_subst_ex_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_subst_ex : lngen.


(* *********************************************************************** *)
(** * Theorems about [fv] *)

(* begin hide *)

Lemma fv_eX_close_eX_rec_mutual :
(forall e X n,
  fv_eX (close_eX_rec n X e) [=] remove X (fv_eX e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_eX_close_eX_rec :
forall e X n,
  fv_eX (close_eX_rec n X e) [=] remove X (fv_eX e).
Proof.
pose proof fv_eX_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_close_eX_rec : lngen.
Hint Rewrite fv_eX_close_eX_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_eX_close_ex_rec_mutual :
(forall e x n,
  fv_eX (close_ex_rec n x e) [=] fv_eX e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_eX_close_ex_rec :
forall e x n,
  fv_eX (close_ex_rec n x e) [=] fv_eX e.
Proof.
pose proof fv_eX_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_close_ex_rec : lngen.
Hint Rewrite fv_eX_close_ex_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_ex_close_eX_rec_mutual :
(forall e X n,
  fv_ex (close_eX_rec n X e) [=] fv_ex e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_ex_close_eX_rec :
forall e X n,
  fv_ex (close_eX_rec n X e) [=] fv_ex e.
Proof.
pose proof fv_ex_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_close_eX_rec : lngen.
Hint Rewrite fv_ex_close_eX_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_ex_close_ex_rec_mutual :
(forall e x n,
  fv_ex (close_ex_rec n x e) [=] remove x (fv_ex e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_ex_close_ex_rec :
forall e x n,
  fv_ex (close_ex_rec n x e) [=] remove x (fv_ex e).
Proof.
pose proof fv_ex_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_close_ex_rec : lngen.
Hint Rewrite fv_ex_close_ex_rec using solve [auto] : lngen.

(* end hide *)

Lemma fv_eX_close_eX :
forall e X,
  fv_eX (close_eX X e) [=] remove X (fv_eX e).
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve fv_eX_close_eX : lngen.
Hint Rewrite fv_eX_close_eX using solve [auto] : lngen.

Lemma fv_eX_close_ex :
forall e x,
  fv_eX (close_ex x e) [=] fv_eX e.
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve fv_eX_close_ex : lngen.
Hint Rewrite fv_eX_close_ex using solve [auto] : lngen.

Lemma fv_ex_close_eX :
forall e X,
  fv_ex (close_eX X e) [=] fv_ex e.
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve fv_ex_close_eX : lngen.
Hint Rewrite fv_ex_close_eX using solve [auto] : lngen.

Lemma fv_ex_close_ex :
forall e x,
  fv_ex (close_ex x e) [=] remove x (fv_ex e).
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve fv_ex_close_ex : lngen.
Hint Rewrite fv_ex_close_ex using solve [auto] : lngen.

(* begin hide *)

Lemma fv_eX_open_eX_rec_lower_mutual :
(forall e e1 n,
  fv_eX e [<=] fv_eX (open_eX_rec n e1 e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_eX_rec_lower :
forall e e1 n,
  fv_eX e [<=] fv_eX (open_eX_rec n e1 e).
Proof.
pose proof fv_eX_open_eX_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_open_eX_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_ex_rec_lower_mutual :
(forall e e1 n,
  fv_eX e [<=] fv_eX (open_ex_rec n e1 e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_ex_rec_lower :
forall e e1 n,
  fv_eX e [<=] fv_eX (open_ex_rec n e1 e).
Proof.
pose proof fv_eX_open_ex_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_open_ex_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_eX_rec_lower_mutual :
(forall e e1 n,
  fv_ex e [<=] fv_ex (open_eX_rec n e1 e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_eX_rec_lower :
forall e e1 n,
  fv_ex e [<=] fv_ex (open_eX_rec n e1 e).
Proof.
pose proof fv_ex_open_eX_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_open_eX_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_ex_rec_lower_mutual :
(forall e e1 n,
  fv_ex e [<=] fv_ex (open_ex_rec n e1 e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_ex_rec_lower :
forall e e1 n,
  fv_ex e [<=] fv_ex (open_ex_rec n e1 e).
Proof.
pose proof fv_ex_open_ex_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_open_ex_rec_lower : lngen.

(* end hide *)

Lemma fv_eX_open_eX_lower :
forall e e1,
  fv_eX e [<=] fv_eX (open_eX e1 e).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve fv_eX_open_eX_lower : lngen.

Lemma fv_eX_open_ex_lower :
forall e e1,
  fv_eX e [<=] fv_eX (open_ex e1 e).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve fv_eX_open_ex_lower : lngen.

Lemma fv_ex_open_eX_lower :
forall e e1,
  fv_ex e [<=] fv_ex (open_eX e1 e).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve fv_ex_open_eX_lower : lngen.

Lemma fv_ex_open_ex_lower :
forall e e1,
  fv_ex e [<=] fv_ex (open_ex e1 e).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve fv_ex_open_ex_lower : lngen.

(* begin hide *)

Lemma fv_eX_open_eX_rec_upper_mutual :
(forall e e1 n,
  fv_eX (open_eX_rec n e1 e) [<=] fv_eX e1 `union` fv_eX e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_eX_rec_upper :
forall e e1 n,
  fv_eX (open_eX_rec n e1 e) [<=] fv_eX e1 `union` fv_eX e.
Proof.
pose proof fv_eX_open_eX_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_open_eX_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_ex_rec_upper_mutual :
(forall e e1 n,
  fv_eX (open_ex_rec n e1 e) [<=] fv_eX e1 `union` fv_eX e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_ex_rec_upper :
forall e e1 n,
  fv_eX (open_ex_rec n e1 e) [<=] fv_eX e1 `union` fv_eX e.
Proof.
pose proof fv_eX_open_ex_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_open_ex_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_eX_rec_upper_mutual :
(forall e e1 n,
  fv_ex (open_eX_rec n e1 e) [<=] fv_ex e1 `union` fv_ex e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_eX_rec_upper :
forall e e1 n,
  fv_ex (open_eX_rec n e1 e) [<=] fv_ex e1 `union` fv_ex e.
Proof.
pose proof fv_ex_open_eX_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_open_eX_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_ex_rec_upper_mutual :
(forall e e1 n,
  fv_ex (open_ex_rec n e1 e) [<=] fv_ex e1 `union` fv_ex e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_ex_rec_upper :
forall e e1 n,
  fv_ex (open_ex_rec n e1 e) [<=] fv_ex e1 `union` fv_ex e.
Proof.
pose proof fv_ex_open_ex_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_open_ex_rec_upper : lngen.

(* end hide *)

Lemma fv_eX_open_eX_upper :
forall e e1,
  fv_eX (open_eX e1 e) [<=] fv_eX e1 `union` fv_eX e.
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve fv_eX_open_eX_upper : lngen.

Lemma fv_eX_open_ex_upper :
forall e e1,
  fv_eX (open_ex e1 e) [<=] fv_eX e1 `union` fv_eX e.
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve fv_eX_open_ex_upper : lngen.

Lemma fv_ex_open_eX_upper :
forall e e1,
  fv_ex (open_eX e1 e) [<=] fv_ex e1 `union` fv_ex e.
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve fv_ex_open_eX_upper : lngen.

Lemma fv_ex_open_ex_upper :
forall e e1,
  fv_ex (open_ex e1 e) [<=] fv_ex e1 `union` fv_ex e.
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve fv_ex_open_ex_upper : lngen.

(* begin hide *)

Lemma fv_eX_subst_eX_fresh_mutual :
(forall e e1 X,
  X `notin` fv_eX e ->
  fv_eX (subst_eX e1 X e) [=] fv_eX e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

Lemma fv_eX_subst_eX_fresh :
forall e e1 X,
  X `notin` fv_eX e ->
  fv_eX (subst_eX e1 X e) [=] fv_eX e.
Proof.
pose proof fv_eX_subst_eX_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_eX_fresh : lngen.
Hint Rewrite fv_eX_subst_eX_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma fv_ex_subst_ex_fresh_mutual :
(forall e e1 x,
  x `notin` fv_ex e ->
  fv_ex (subst_ex e1 x e) [=] fv_ex e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

Lemma fv_ex_subst_ex_fresh :
forall e e1 x,
  x `notin` fv_ex e ->
  fv_ex (subst_ex e1 x e) [=] fv_ex e.
Proof.
pose proof fv_ex_subst_ex_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_ex_fresh : lngen.
Hint Rewrite fv_ex_subst_ex_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma fv_eX_subst_eX_lower_mutual :
(forall e e1 X,
  remove X (fv_eX e) [<=] fv_eX (subst_eX e1 X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

Lemma fv_eX_subst_eX_lower :
forall e e1 X,
  remove X (fv_eX e) [<=] fv_eX (subst_eX e1 X e).
Proof.
pose proof fv_eX_subst_eX_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_eX_lower : lngen.

(* begin hide *)

Lemma fv_eX_subst_ex_lower_mutual :
(forall e e1 x,
  fv_eX e [<=] fv_eX (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

Lemma fv_eX_subst_ex_lower :
forall e e1 x,
  fv_eX e [<=] fv_eX (subst_ex e1 x e).
Proof.
pose proof fv_eX_subst_ex_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_ex_lower : lngen.

(* begin hide *)

Lemma fv_ex_subst_eX_lower_mutual :
(forall e e1 X,
  fv_ex e [<=] fv_ex (subst_eX e1 X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

Lemma fv_ex_subst_eX_lower :
forall e e1 X,
  fv_ex e [<=] fv_ex (subst_eX e1 X e).
Proof.
pose proof fv_ex_subst_eX_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_eX_lower : lngen.

(* begin hide *)

Lemma fv_ex_subst_ex_lower_mutual :
(forall e e1 x,
  remove x (fv_ex e) [<=] fv_ex (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

Lemma fv_ex_subst_ex_lower :
forall e e1 x,
  remove x (fv_ex e) [<=] fv_ex (subst_ex e1 x e).
Proof.
pose proof fv_ex_subst_ex_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_ex_lower : lngen.

(* begin hide *)

Lemma fv_eX_subst_eX_notin_mutual :
(forall e e1 X X1,
  X1 `notin` fv_eX e ->
  X1 `notin` fv_eX e1 ->
  X1 `notin` fv_eX (subst_eX e1 X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

Lemma fv_eX_subst_eX_notin :
forall e e1 X X1,
  X1 `notin` fv_eX e ->
  X1 `notin` fv_eX e1 ->
  X1 `notin` fv_eX (subst_eX e1 X e).
Proof.
pose proof fv_eX_subst_eX_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_eX_notin : lngen.

(* begin hide *)

Lemma fv_eX_subst_ex_notin_mutual :
(forall e e1 x X,
  X `notin` fv_eX e ->
  X `notin` fv_eX e1 ->
  X `notin` fv_eX (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

Lemma fv_eX_subst_ex_notin :
forall e e1 x X,
  X `notin` fv_eX e ->
  X `notin` fv_eX e1 ->
  X `notin` fv_eX (subst_ex e1 x e).
Proof.
pose proof fv_eX_subst_ex_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_ex_notin : lngen.

(* begin hide *)

Lemma fv_ex_subst_eX_notin_mutual :
(forall e e1 X x,
  x `notin` fv_ex e ->
  x `notin` fv_ex e1 ->
  x `notin` fv_ex (subst_eX e1 X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

Lemma fv_ex_subst_eX_notin :
forall e e1 X x,
  x `notin` fv_ex e ->
  x `notin` fv_ex e1 ->
  x `notin` fv_ex (subst_eX e1 X e).
Proof.
pose proof fv_ex_subst_eX_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_eX_notin : lngen.

(* begin hide *)

Lemma fv_ex_subst_ex_notin_mutual :
(forall e e1 x x1,
  x1 `notin` fv_ex e ->
  x1 `notin` fv_ex e1 ->
  x1 `notin` fv_ex (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

Lemma fv_ex_subst_ex_notin :
forall e e1 x x1,
  x1 `notin` fv_ex e ->
  x1 `notin` fv_ex e1 ->
  x1 `notin` fv_ex (subst_ex e1 x e).
Proof.
pose proof fv_ex_subst_ex_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_ex_notin : lngen.

(* begin hide *)

Lemma fv_eX_subst_eX_upper_mutual :
(forall e e1 X,
  fv_eX (subst_eX e1 X e) [<=] fv_eX e1 `union` remove X (fv_eX e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

Lemma fv_eX_subst_eX_upper :
forall e e1 X,
  fv_eX (subst_eX e1 X e) [<=] fv_eX e1 `union` remove X (fv_eX e).
Proof.
pose proof fv_eX_subst_eX_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_eX_upper : lngen.

(* begin hide *)

Lemma fv_eX_subst_ex_upper_mutual :
(forall e e1 x,
  fv_eX (subst_ex e1 x e) [<=] fv_eX e1 `union` fv_eX e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec another_atom.
Qed.

(* end hide *)

Lemma fv_eX_subst_ex_upper :
forall e e1 x,
  fv_eX (subst_ex e1 x e) [<=] fv_eX e1 `union` fv_eX e.
Proof.
pose proof fv_eX_subst_ex_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_ex_upper : lngen.

(* begin hide *)

Lemma fv_ex_subst_eX_upper_mutual :
(forall e e1 X,
  fv_ex (subst_eX e1 X e) [<=] fv_ex e1 `union` fv_ex e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

Lemma fv_ex_subst_eX_upper :
forall e e1 X,
  fv_ex (subst_eX e1 X e) [<=] fv_ex e1 `union` fv_ex e.
Proof.
pose proof fv_ex_subst_eX_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_eX_upper : lngen.

(* begin hide *)

Lemma fv_ex_subst_ex_upper_mutual :
(forall e e1 x,
  fv_ex (subst_ex e1 x e) [<=] fv_ex e1 `union` remove x (fv_ex e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec atom.
Qed.

(* end hide *)

Lemma fv_ex_subst_ex_upper :
forall e e1 x,
  fv_ex (subst_ex e1 x e) [<=] fv_ex e1 `union` remove x (fv_ex e).
Proof.
pose proof fv_ex_subst_ex_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_ex_upper : lngen.


(* *********************************************************************** *)
(** * More theorems about [subst] *)

(* begin hide *)

Lemma subst_eX_close_eX_rec_open_eX_rec_mutual :
(forall e1 e X X1 n,
  X1 `notin` fv_eX e1 ->
  X1 `notin` fv_eX e ->
  X1 <> X ->
  degree_eX n e ->
  subst_eX e X e1 = close_eX_rec n X1 (subst_eX e X (open_eX_rec n (xvar X1) e1))).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_eX_close_eX_rec_open_eX_rec :
forall e1 e X X1 n,
  X1 `notin` fv_eX e1 ->
  X1 `notin` fv_eX e ->
  X1 <> X ->
  degree_eX n e ->
  subst_eX e X e1 = close_eX_rec n X1 (subst_eX e X (open_eX_rec n (xvar X1) e1)).
Proof.
pose proof subst_eX_close_eX_rec_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_close_eX_rec_open_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_eX_close_ex_rec_open_ex_rec_mutual :
(forall e1 e X x n,
  x `notin` fv_ex e1 ->
  x `notin` fv_ex e ->
  degree_ex n e ->
  subst_eX e X e1 = close_ex_rec n x (subst_eX e X (open_ex_rec n (var x) e1))).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_eX_close_ex_rec_open_ex_rec :
forall e1 e X x n,
  x `notin` fv_ex e1 ->
  x `notin` fv_ex e ->
  degree_ex n e ->
  subst_eX e X e1 = close_ex_rec n x (subst_eX e X (open_ex_rec n (var x) e1)).
Proof.
pose proof subst_eX_close_ex_rec_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_close_ex_rec_open_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_ex_close_eX_rec_open_eX_rec_mutual :
(forall e1 e x X n,
  X `notin` fv_eX e1 ->
  X `notin` fv_eX e ->
  degree_eX n e ->
  subst_ex e x e1 = close_eX_rec n X (subst_ex e x (open_eX_rec n (xvar X) e1))).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_ex_close_eX_rec_open_eX_rec :
forall e1 e x X n,
  X `notin` fv_eX e1 ->
  X `notin` fv_eX e ->
  degree_eX n e ->
  subst_ex e x e1 = close_eX_rec n X (subst_ex e x (open_eX_rec n (xvar X) e1)).
Proof.
pose proof subst_ex_close_eX_rec_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_close_eX_rec_open_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_ex_close_ex_rec_open_ex_rec_mutual :
(forall e1 e x x1 n,
  x1 `notin` fv_ex e1 ->
  x1 `notin` fv_ex e ->
  x1 <> x ->
  degree_ex n e ->
  subst_ex e x e1 = close_ex_rec n x1 (subst_ex e x (open_ex_rec n (var x1) e1))).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_ex_close_ex_rec_open_ex_rec :
forall e1 e x x1 n,
  x1 `notin` fv_ex e1 ->
  x1 `notin` fv_ex e ->
  x1 <> x ->
  degree_ex n e ->
  subst_ex e x e1 = close_ex_rec n x1 (subst_ex e x (open_ex_rec n (var x1) e1)).
Proof.
pose proof subst_ex_close_ex_rec_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_close_ex_rec_open_ex_rec : lngen.

(* end hide *)

Lemma subst_eX_close_eX_open_eX :
forall e1 e X X1,
  X1 `notin` fv_eX e1 ->
  X1 `notin` fv_eX e ->
  X1 <> X ->
  lc_e e ->
  subst_eX e X e1 = close_eX X1 (subst_eX e X (open_eX (xvar X1) e1)).
Proof.
unfold close_eX; unfold open_eX; default_simp.
Qed.

Hint Resolve subst_eX_close_eX_open_eX : lngen.

Lemma subst_eX_close_ex_open_ex :
forall e1 e X x,
  x `notin` fv_ex e1 ->
  x `notin` fv_ex e ->
  lc_e e ->
  subst_eX e X e1 = close_ex x (subst_eX e X (open_ex (var x) e1)).
Proof.
unfold close_ex; unfold open_ex; default_simp.
Qed.

Hint Resolve subst_eX_close_ex_open_ex : lngen.

Lemma subst_ex_close_eX_open_eX :
forall e1 e x X,
  X `notin` fv_eX e1 ->
  X `notin` fv_eX e ->
  lc_e e ->
  subst_ex e x e1 = close_eX X (subst_ex e x (open_eX (xvar X) e1)).
Proof.
unfold close_eX; unfold open_eX; default_simp.
Qed.

Hint Resolve subst_ex_close_eX_open_eX : lngen.

Lemma subst_ex_close_ex_open_ex :
forall e1 e x x1,
  x1 `notin` fv_ex e1 ->
  x1 `notin` fv_ex e ->
  x1 <> x ->
  lc_e e ->
  subst_ex e x e1 = close_ex x1 (subst_ex e x (open_ex (var x1) e1)).
Proof.
unfold close_ex; unfold open_ex; default_simp.
Qed.

Hint Resolve subst_ex_close_ex_open_ex : lngen.

Lemma subst_eX_abs :
forall x e1 e X,
  lc_e e ->
  x `notin` fv_ex e `union` fv_ex e1 ->
  subst_eX e X (abs e1) = abs (close_ex x (subst_eX e X (open_ex (var x) e1))).
Proof.
default_simp.
Qed.

Hint Resolve subst_eX_abs : lngen.

Lemma subst_eX_xabs :
forall X1 e1 e X,
  lc_e e ->
  X1 `notin` fv_eX e `union` fv_eX e1 `union` singleton X ->
  subst_eX e X (xabs e1) = xabs (close_eX X1 (subst_eX e X (open_eX (xvar X1) e1))).
Proof.
default_simp.
Qed.

Hint Resolve subst_eX_xabs : lngen.

Lemma subst_ex_abs :
forall x1 e1 e x,
  lc_e e ->
  x1 `notin` fv_ex e `union` fv_ex e1 `union` singleton x ->
  subst_ex e x (abs e1) = abs (close_ex x1 (subst_ex e x (open_ex (var x1) e1))).
Proof.
default_simp.
Qed.

Hint Resolve subst_ex_abs : lngen.

Lemma subst_ex_xabs :
forall X e1 e x,
  lc_e e ->
  X `notin` fv_eX e `union` fv_eX e1 ->
  subst_ex e x (xabs e1) = xabs (close_eX X (subst_ex e x (open_eX (xvar X) e1))).
Proof.
default_simp.
Qed.

Hint Resolve subst_ex_xabs : lngen.

